import socket
from unittest import mock

import pytest
from bec_server.device_server.tests.utils import DMMock

from ophyd_devices.tests.utils import SocketMock
from ophyd_devices.utils.controller import Controller
from ophyd_devices.utils.socket import SocketIO, SocketSignal


class DummySocketSignal(SocketSignal):
    """Dummy SocketSignal class for testing the SocketSignal interface."""

    def __init__(self, name, controller: Controller, **kwargs):
        super().__init__(name=name, **kwargs)
        self.controller = controller

    def _socket_get(self) -> str:
        return self.controller.socket_put_and_receive("get")

    def _socket_set(self, value: str):
        self.controller.socket_put_and_receive(value)


@pytest.fixture
def controller():
    """Controller fixture for testing the SocketSignal interface."""
    try:
        dm = DMMock()
        Controller._reset_controller()
        controller = Controller(
            name="controller",
            socket_cls=SocketMock,
            socket_host="localhost",
            socket_port=8080,
            device_manager=dm,
        )
        controller.on()
        yield controller
    finally:
        Controller._reset_controller()


@pytest.fixture
def signal(controller):
    """Dummy SocketSignal fixture for testing."""
    return DummySocketSignal(name="signal", auto_monitor=True, controller=controller)


class DummySocket:
    AF_INET = 2
    SOCK_STREAM = 1

    def __init__(self) -> None:
        self.address_family = None
        self.socket_kind = None
        self.timeout = None

    def socket(self, address_family, socket_kind):
        self.address_family = address_family
        self.socket_kind = socket_kind
        return self

    def settimeout(self, timeout):
        self.timeout = timeout

    def send(self, msg, *args, **kwargs):
        self.send_buffer = msg

    def connect(self, address):
        self.host = address[0]
        self.port = address[1]
        self.connected = True

    def close(self):
        self.connected = False


def test_socket_init():
    socketio = SocketIO("localhost", 8080)

    assert socketio.host == "localhost"
    assert socketio.port == 8080

    assert socketio.is_open == False

    assert socketio.sock.family == socket.AF_INET
    assert socketio.sock.type == socket.SOCK_STREAM


def test_socket_put():
    dsocket = DummySocket()
    socketio = SocketIO("localhost", 8080)
    socketio.sock = dsocket
    socketio.put(b"message")
    assert dsocket.send_buffer == b"message"


def test_open():
    dsocket = DummySocket()
    socketio = SocketIO("localhost", 8080)
    socketio.sock = dsocket
    socketio.open()
    assert socketio.is_open == True
    assert socketio.sock.host == socketio.host
    assert socketio.sock.port == socketio.port


def test_socket_open_with_timeout():
    dsocket = DummySocket()
    socketio = SocketIO("localhost", 8080)
    socketio.sock = dsocket
    with mock.patch.object(dsocket, "connect") as mock_connect:
        socketio.open(timeout=0.1)
        mock_connect.assert_called_once()
        mock_connect.reset_mock()
        # There is a 1s sleep in the retry loop, mock_connect should be called only once
        mock_connect.side_effect = Exception("Connection failed")
        with pytest.raises(ConnectionError):
            socketio.open(timeout=0.4)
        mock_connect.assert_called_once()


def test_close():
    socketio = SocketIO("localhost", 8080)
    socketio.close()
    assert socketio.sock == None
    assert socketio.is_open == False


def test_socket_signal_get(signal):
    """
    Test that the callback mechanism of SocketSignal correctly handles recursive reads without
    causing multiple socket reads, and that the value is correctly cached and passed to the callback.
    """
    controller = signal.controller
    assert signal._auto_monitor == True

    controller.sock: SocketMock
    controller.sock.buffer_recv = [b"value2", b"value1"]

    callback_read_buffer = []
    callback_value_buffer = []

    readback = signal.read()
    assert readback[signal.name]["value"] == "value2"

    def _test_cb(value, old_value, **kwargs):
        """Simulate a callback that triggers another read."""
        signal = kwargs["obj"]
        callback_value_buffer.append((value, old_value))
        readback = signal.read()
        callback_read_buffer.append(readback)

    signal.subscribe(_test_cb, event_type=signal.SUB_VALUE, run=False)

    readback = signal.read()
    assert len(callback_read_buffer) == 1, "Callback should have been called once"
    assert len(callback_value_buffer) == 1, "Callback should have been called once"

    assert readback == callback_read_buffer[0]
    assert callback_value_buffer == [("value1", "value2")]


def test_socket_signal_put(signal):
    """
    Test that the put method of the SocketSignal class correctly sends values to the socket,
    and that it implements the necessary subscription notifications for value changes.
    """
    controller = signal.controller
    controller.sock: SocketMock

    controller.sock.buffer_recv = [b"value2", b"new_value", b"new_value"]

    callback_read_buffer = []
    callback_value_buffer = []

    readback = signal.read()
    assert readback[signal.name]["value"] == "value2"

    def _test_readback_cb(value, old_value, **kwargs):
        """Simulate a callback that triggers another read."""
        signal = kwargs["obj"]
        callback_value_buffer.append((value, old_value))
        readback = signal.read()
        callback_read_buffer.append(readback)

    callback_setpoint_buffer = []

    def _test_setpoint_cb(value, old_value, **kwargs):
        """Simulate a callback that runs a read, this should trigger another read on the socket."""
        signal = kwargs["obj"]
        callback_setpoint_buffer.append((value, old_value))
        signal.read()

    signal.subscribe(_test_setpoint_cb, event_type=signal.SUB_SETPOINT, run=False)
    signal.subscribe(_test_readback_cb, event_type=signal.SUB_VALUE, run=False)

    # Now we run 'put'. This runs super().put(...), which triggers the SUB_VALUE callback first,
    # and then the SUB_SETPOINT callback. With our extra callback on SUB_VALUE, we will trigger
    # one _test_readback_cb callback through the super().put(...) call with sub_type SUB_VALUE,
    # and another _test_readback_cb callback through the _test_setpoint_cb callback with sub_type SUB_SETPOINT.
    # Respectively the signal.read() call in test_setpoint_cb.
    signal.put("new_value")
    assert controller.sock.buffer_put == [b"get\n", b"new_value\n", b"get\n"]
    assert len(callback_setpoint_buffer) == 1, "Setpoint callback should have been called once"
    assert len(callback_read_buffer) == 2, "Readback callback should have been called twice"
    assert len(callback_value_buffer) == 2, "Value callback should have been called twice"
    assert callback_setpoint_buffer == [("new_value", "value2")]
    assert callback_read_buffer[0][signal.name]["value"] == "new_value"
    assert callback_read_buffer[1][signal.name]["value"] == "new_value"
    assert callback_value_buffer == [("new_value", "value2"), ("new_value", "new_value")]



"""Utility handler to run tasks (function, conditions) in an asynchronous fashion."""

import ctypes
import operator
import threading
import traceback
import uuid
from enum import Enum
from typing import TYPE_CHECKING, Callable, Literal, Union

from bec_lib.file_utils import get_full_path
from bec_lib.logger import bec_logger
from bec_lib.utils.import_utils import lazy_import_from
from ophyd.status import DeviceStatus as _DeviceStatus
from ophyd.status import MoveStatus as _MoveStatus
from ophyd.status import Status as _Status
from ophyd.status import StatusBase as _StatusBase

if TYPE_CHECKING:  # pragma: no cover
    from bec_lib.messages import ScanStatusMessage
    from ophyd import Device, Signal
else:
    # TODO: put back normal import when Pydantic gets faster
    ScanStatusMessage = lazy_import_from("bec_lib.messages", ("ScanStatusMessage",))


__all__ = [
    "CompareStatus",
    "TransitionStatus",
    "AndStatus",
    "DeviceStatus",
    "MoveStatus",
    "Status",
    "StatusBase",
    "SubscriptionStatus",
]

logger = bec_logger.logger

set_async_exc = ctypes.pythonapi.PyThreadState_SetAsyncExc

OP_MAP = {
    "==": operator.eq,
    "!=": operator.ne,
    "<": operator.lt,
    "<=": operator.le,
    ">": operator.gt,
    ">=": operator.ge,
}


class StatusBase(_StatusBase):
    """Base class for all status objects."""

    def __init__(
        self,
        obj: Union["Device", None] = None,
        *,
        timeout=None,
        settle_time=0,
        done=None,
        success=None,
    ):
        self.obj = obj
        super().__init__(timeout=timeout, settle_time=settle_time, done=done, success=success)

    def __and__(self, other):
        """Returns a new 'composite' status object, AndStatus"""
        return AndStatus(self, other)


class AndStatus(StatusBase):
    """
    A Status that has composes two other Status objects using logical and.
    If any of the two Status objects fails, the combined status will fail
    with the exception of the first Status to fail.

    Args:
        left (StatusBase): Left status object.
        right (StatusBase): Right status object.

    Examples:
        >>> status1 = StatusBase(device1)
        >>> status2 = StatusBase(device2)
        >>> status3 = StatusBase(device3)
        >>> combined_status = AndStatus(status1, status2)
        >>> combined_status = status1 & status2
        >>> combined_status = status1 & status2 & status3
    """

    def __init__(self, left, right, **kwargs):
        self.left = left
        self.right = right
        super().__init__(**kwargs)
        self._trace_attributes["left"] = self.left._trace_attributes
        self._trace_attributes["right"] = self.right._trace_attributes

        def inner(status):
            with self._lock:
                if self._externally_initiated_completion:
                    return

                # Return if status is already done..
                if self.done:
                    return

                with status._lock:
                    if status.done and not status.success:
                        self.set_exception(status.exception())  # st._exception
                        return
                if self.left.done and self.right.done and self.left.success and self.right.success:
                    self.set_finished()

        self.left.add_callback(inner)
        self.right.add_callback(inner)

    def __repr__(self):
        return "({self.left!r} & {self.right!r})".format(self=self)

    def __str__(self):
        return "{0}(done={1.done}, " "success={1.success})" "".format(self.__class__.__name__, self)

    def __contains__(self, status) -> bool:
        for child in [self.left, self.right]:
            if child == status:
                return True
            if isinstance(child, AndStatus):
                if status in child:
                    return True

        return False


class Status(_Status):
    """Thin wrapper around StatusBase to add __and__ operator."""

    def __and__(self, other):
        """Returns a new 'composite' status object, AndStatus"""
        return AndStatus(self, other)


class DeviceStatus(_DeviceStatus):
    """Thin wrapper around DeviceStatus to add __and__ operator."""

    def __and__(self, other):
        """Returns a new 'composite' status object, AndStatus"""
        return AndStatus(self, other)


class MoveStatus(_MoveStatus):
    """Thin wrapper around MoveStatus to ensure __and__ operator and stop on failure."""

    def __and__(self, other):
        """Returns a new 'composite' status object, AndStatus"""
        return AndStatus(self, other)


class SubscriptionStatus(StatusBase):
    """Subscription status implementation based on wrapped StatusBase implementation."""

    def __init__(
        self,
        obj: Union["Device", "Signal"],
        callback: Callable,
        event_type=None,
        timeout=None,
        settle_time=None,
        run=True,
    ):
        # Store device and attribute information
        self.callback = callback
        self.obj = obj
        # Start timeout thread in the background
        super().__init__(obj=obj, timeout=timeout, settle_time=settle_time)

        self.obj.subscribe(self.check_value, event_type=event_type, run=run)

    def check_value(self, *args, **kwargs):
        """Update the status object"""
        try:
            success = self.callback(*args, **kwargs)
        except Exception as e:
            logger.error(f"Error in SubscriptionStatus callback: {e}")
            self.set_exception(e)
            return
        if success:
            self.set_finished()

    def set_finished(self):
        """Mark as finished successfully."""
        self.obj.clear_sub(self.check_value)
        super().set_finished()

    def _handle_failure(self):
        """Clear subscription on failure, run callbacks through super()"""
        self.obj.clear_sub(self.check_value)
        return super()._handle_failure()


class CompareStatus(SubscriptionStatus):
    """
    Status to compare a signal value against a given value.
    The comparison is done using the specified operation, which can be one of
    '==', '!=', '<', '<=', '>', '>='. If the value is a string, only '==' and '!=' are allowed.
    One may also define a value or list of values that will result in an exception if encountered.
    The status is finished when the comparison is either true or an exception is raised.

    Args:
        signal (Signal): The signal to monitor.
        value (float | int | str): The target value to compare against.
        operation_success (str, optional): The comparison operation for success. Defaults to '=='.
        failure_value (float | int | str | list[float | int | str] | None, optional):
            A value or list of values that will trigger an exception if encountered. Defaults to None.
        operation_failure (str, optional): The comparison operation for failure values. Defaults to '=='.
        event_type (int, optional): The event type to subscribe to. Defaults to None.
        timeout (float, optional): Timeout for the status. Defaults to None.
        settle_time (float, optional): Settle time before checking the status. Defaults to 0.
        run (bool, optional): Whether to start the status immediately. Defaults to True
    """

    def __init__(
        self,
        signal: "Signal",
        value: float | int | str,
        *,
        operation_success: Literal["==", "!=", "<", "<=", ">", ">="] = "==",
        failure_value: float | int | str | list[float | int | str] | None = None,
        operation_failure: Literal["==", "!=", "<", "<=", ">", ">="] = "==",
        timeout: float = None,
        settle_time: float = 0,
        run: bool = True,
        event_type=None,
    ):
        if isinstance(value, str):
            if operation_success not in ("==", "!=") or operation_failure not in ("==", "!="):
                raise ValueError(
                    f"Invalid operation_success: {operation_success} for string comparison. Must be '==' or '!='."
                )
        if operation_success not in ("==", "!=", "<", "<=", ">", ">="):
            raise ValueError(
                f"Invalid operation_success: {operation_success}. Must be one of '==', '!=', '<', '<=', '>', '>='."
            )
        self._signal = signal
        self._value = value
        self._operation_success = operation_success
        self._operation_failure = operation_failure
        self.op_map = OP_MAP
        if failure_value is None:
            self._failure_values = []
        elif isinstance(failure_value, (float, int, str)):
            self._failure_values = [failure_value]
        elif isinstance(failure_value, (list, tuple)):
            self._failure_values = failure_value
        else:
            raise ValueError(
                f"failure_value must be a float, int, str, list or None. Received: {failure_value}"
            )
        super().__init__(
            obj=signal,
            callback=self._compare_callback,
            timeout=timeout,
            settle_time=settle_time,
            event_type=event_type,
            run=run,
        )

    def _compare_callback(self, value: any, **kwargs) -> bool:
        """
        Callback for subscription status

        Args:
            value (any): Current value of the signal

        Returns:
            bool: True if comparison is successful, False otherwise.
        """
        try:
            if isinstance(value, list):
                raise ValueError(f"List values are not supported. Received value: {value}")
            if any(
                self.op_map[self._operation_failure](value, failure_value)
                for failure_value in self._failure_values
            ):
                raise ValueError(
                    f"CompareStatus for signal {self._signal.name} "
                    f"did not reach the desired state {self._operation_success} {self._value}. "
                    f"But instead reached {value}, which is in list of failure values: {self._failure_values}"
                )
            return self.op_map[self._operation_success](value, self._value)
        except Exception as e:
            logger.error(f"Error in CompareStatus callback: {e}")
            self.set_exception(e)
            return False


class TransitionStatus(SubscriptionStatus):
    """
    Status to monitor transitions of a signal value through a list of specified transitions.
    The status is finished when all transitions have been observed in order. The keyword argument
    `strict` determines whether the transitions must occur in strict order or not. The strict option
    only becomes relevant once the first transition has been observed.
    If `failure_states` is provided, the status will raise an exception if the signal value matches
    any of the values in `failure_states`.

    Args:
        signal (Signal): The signal to monitor.
        transitions (list[float | int | str]): List of values representing the transitions to observe.
        strict (bool, optional): Whether to enforce strict order of transitions. Defaults to True.
        failure_states (list[float | int | str] | None, optional):
            A list of values that will trigger an exception if encountered. Defaults to None.
        run (bool, optional): Whether to start the status immediately. Defaults to True.
        event_type (int, optional): The event type to subscribe to. Defaults to None.
        timeout (float, optional): Timeout for the status. Defaults to None.
        settle_time (float, optional): Settle time before checking the status. Defaults to 0.

    Notes:
        The 'strict' option does not raise if transitions are observed which are out of order.
        It only determines whether a transition is accepted if it is observed from the
        previous value in the list of transitions to the next value.
        For example, with strict=True and transitions=[1, 2, 3], the sequence
        0 -> 1 -> 2 -> 3 is accepted, but 0 -> 1 -> 3 -> 2 -> 3 is not and the status
        will not complete. With strict=False, both sequences are accepted.
        However, with strict=True, the sequence 0 -> 1 -> 3 -> 1 -> 2 -> 3 is accepted.
        To raise an exception if an out-of-order transition is observed, use the
        `failure_states` keyword argument.
    """

    def __init__(
        self,
        signal: "Signal",
        transitions: list[float | int | str],
        *,
        strict: bool = True,
        failure_states: list[float | int | str] | None = None,
        run: bool = True,
        timeout: float = None,
        settle_time: float = 0,
        event_type=None,
    ):
        self._signal = signal
        self._transitions = tuple(transitions)
        if not transitions:
            raise ValueError("Transitions {transitions}must contain at least one value")
        self._index = 0
        self._strict = strict
        self._failure_states = failure_states if failure_states else []
        super().__init__(
            obj=signal,
            callback=self._compare_callback,
            timeout=timeout,
            settle_time=settle_time,
            event_type=event_type,
            run=run,
        )

    def _compare_callback(self, old_value: any, value: any, **kwargs) -> bool:
        """
        Callback for subscription Status

        Args:
            old_value (any): Previous value of the signal
            value (any): Current value of the signal

        Returns:
            bool: True if all transitions have been observed, False otherwise.
        """
        try:
            if value in self._failure_states:
                raise ValueError(
                    f"Transition Status for {self._signal.name} resulted in a value: {value}. "
                    f"marked to raise {self._failure_states}. Expected transitions: {self._transitions}."
                )
            if self._index == 0:
                if value == self._transitions[0]:
                    self._index += 1
            else:
                if self._strict:
                    if (
                        old_value == self._transitions[self._index - 1]
                        and value == self._transitions[self._index]
                    ):
                        self._index += 1
                else:
                    if value == self._transitions[self._index]:
                        self._index += 1
            return self._index >= len(self._transitions)
        except Exception as e:
            # Catch any exception if the value comparison fails, e.g. value is numpy array
            logger.error(f"Error in TransitionStatus callback: {e}")
            self.set_exception(e)
            return False


class TaskState(str, Enum):
    """Possible task states"""

    NOT_STARTED = "not_started"
    RUNNING = "running"
    TIMEOUT = "timeout"
    ERROR = "error"
    COMPLETED = "completed"
    KILLED = "killed"


class TaskKilledError(Exception):
    """Exception raised when a task thread is killed"""


class TaskStatus(StatusBase):
    """Thin wrapper around StatusBase to add information about tasks"""

    def __init__(
        self,
        obj: Union["Device", "Signal"],
        *,
        timeout=None,
        settle_time=0,
        done=None,
        success=None,
    ):
        super().__init__(
            obj=obj, timeout=timeout, settle_time=settle_time, done=done, success=success
        )
        self._state = TaskState.NOT_STARTED
        self._task_id = str(uuid.uuid4())

    @property
    def state(self) -> str:
        """Get the state of the task"""
        return self._state.value

    @state.setter
    def state(self, value: TaskState):
        self._state = TaskState(value)

    @property
    def task_id(self) -> str:
        """Get the task ID"""
        return self._task_id


class TaskHandler:
    """Handler to manage asynchronous tasks"""

    def __init__(self, parent: "Device"):
        """Initialize the handler"""
        self._tasks = {}
        self._parent = parent
        self._lock = threading.RLock()

    def submit_task(
        self,
        task: Callable,
        task_args: tuple | None = None,
        task_kwargs: dict | None = None,
        run: bool = True,
    ) -> TaskStatus:
        """Submit a task to the task handler.

        Args:
            task: The task to run.
            run: Whether to run the task immediately.
        """
        task_args = task_args if task_args else ()
        task_kwargs = task_kwargs if task_kwargs else {}
        task_status = TaskStatus(self._parent)
        thread = threading.Thread(
            target=self._wrap_task,
            args=(task, task_args, task_kwargs, task_status),
            name=f"task {task_status.task_id}",
            daemon=True,
        )
        self._tasks.update({task_status.task_id: (task_status, thread)})
        if run is True:
            self.start_task(task_status)
        return task_status

    def start_task(self, task_status: TaskStatus) -> None:
        """Start a task,

        Args:
            task_status: The task status object.
        """
        thread = self._tasks[task_status.task_id][1]
        if thread.is_alive():
            logger.warning(f"Task with ID {task_status.task_id} is already running.")
            return
        task_status.state = TaskState.RUNNING
        thread.start()

    def _wrap_task(
        self, task: Callable, task_args: tuple, task_kwargs: dict, task_status: TaskStatus
    ):
        """Wrap the task in a function"""
        try:
            task(*task_args, **task_kwargs)
        except TimeoutError as exc:
            content = traceback.format_exc()
            logger.warning(
                (
                    f"Timeout Exception in task handler for task {task_status.task_id},"
                    f" Traceback: {content}"
                )
            )
            task_status.state = TaskState.TIMEOUT
            task_status.set_exception(exc)
        except TaskKilledError as exc:
            exc = exc.__class__(
                f"Task {task_status.task_id} was killed. ThreadID:"
                f" {self._tasks[task_status.task_id][1].ident}"
            )
            content = traceback.format_exc()
            logger.warning(
                (
                    f"TaskKilled Exception in task handler for task {task_status.task_id},"
                    f" Traceback: {content}"
                )
            )
            task_status.state = TaskState.KILLED
            task_status.set_exception(exc)
        except Exception as exc:  # pylint: disable=broad-except
            content = traceback.format_exc()
            logger.warning(
                f"Exception in task handler for task {task_status.task_id}, Traceback: {content}"
            )
            task_status.state = TaskState.ERROR
            task_status.set_exception(exc)
        else:
            task_status.state = TaskState.COMPLETED
            task_status.set_finished()
        finally:
            with self._lock:
                self._tasks.pop(task_status.task_id, None)

    def kill_task(self, task_status: TaskStatus) -> None:
        """Kill the thread

        task_status: The task status object.
        """
        thread = self._tasks[task_status.task_id][1]
        exception_cls = TaskKilledError

        ident = ctypes.c_long(thread.ident)
        exc = ctypes.py_object(exception_cls)
        try:
            res = set_async_exc(ident, exc)
            if res == 0:
                raise ValueError("Invalid thread ID")
            if res > 1:
                set_async_exc(ident, None)
                logger.warning(f"Exception raise while kille Thread {ident}; return value: {res}")
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(f"Exception raised while killing thread {ident}: {e}")

    def shutdown(self):
        """Shutdown all tasks of task handler"""
        with self._lock:
            for info in self._tasks.values():
                self.kill_task(info[0])


class FileHandler:
    """Utility class for file operations."""

    def get_full_path(
        self, scan_status_msg: ScanStatusMessage, name: str, create_dir: bool = True
    ) -> str:
        """Get the file path.

        Args:
            scan_info_msg: The scan info message.
            name: The name of the file.
            create_dir: Whether to create the directory.
        """
        return get_full_path(scan_status_msg=scan_status_msg, name=name, create_dir=create_dir)



import numpy as np
from scipy.ndimage import gaussian_filter

from ophyd_devices.sim.sim_data import NoiseType
from ophyd_devices.sim.sim_frameworks.device_proxy import DeviceProxy


class SlitProxy(DeviceProxy):
    """
    Simulation framework to imitate the behaviour of slits.

    This device is a proxy that is meant to override the behaviour of a SimCamera.
    You may use this to simulate the effect of slits on the camera image.

    Parameters can be configured via the DeviceConfig field in the device_config.
    The example below shows the configuration for a pinhole simulation on an Eiger detector,
    where the pinhole is defined by the position of motors samx and samy. These devices must
    exist in your config.

    To update for instance the pixel_size directly, you can directly access the DeviceConfig via
    `dev.eiger.get_device_config()` or update it `dev.eiger.set_device_config({'eiger' : {'pixel_size': 0.1}})`

    An example for the configuration of this is device is in ophyd_devices.configs.ophyd_devices_simulation.yaml
    """

    USER_ACCESS = ["enabled", "lookup", "help"]

    def __init__(self, name, *args, device_manager=None, **kwargs):
        self._gaussian_blur_sigma = 5
        super().__init__(name, *args, device_manager=device_manager, **kwargs)

    def help(self) -> None:
        """Print documentation for the SlitLookup device."""
        print(self.__doc__)

    def _compute(self, device_name: str, *args, **kwargs) -> np.ndarray:
        """
        Compute the lookup table for the simulated camera.
        It copies the sim_camera behaviour and adds a mask to simulate the effect of a pinhole.

        Args:
            device_name (str): Name of the device.
            signal_name (str): Name of the signal.

        Returns:
            np.ndarray: Lookup table for the simulated camera.
        """
        device_obj = self.device_manager.devices.get(device_name).obj
        params = device_obj.sim.params
        shape = device_obj.image_shape.get()
        params.update(
            {
                "noise": NoiseType.POISSON,
                "covariance": np.array(self.config[device_name]["covariance"]),
                "center_offset": np.array(self.config[device_name]["center_offset"]),
            }
        )
        amp = params.get("amplitude")
        cov = params.get("covariance")
        cen_off = params.get("center_offset")

        pos, offset, cov, amp = device_obj.sim._prepare_params_gauss(
            amp=amp, cov=cov, offset=cen_off, shape=shape
        )
        v = device_obj.sim._compute_multivariate_gaussian(pos=pos, cen_off=offset, cov=cov, amp=amp)
        device_pos = self.config[device_name]["pixel_size"] * pos
        valid_mask = self._create_mask(
            device_pos=device_pos,
            ref_motors=self.config[device_name]["ref_motors"],
            width=self.config[device_name]["slit_width"],
            direction=self.config[device_name]["motor_dir"],
        )
        valid_mask = self._blur_image(valid_mask, sigma=self._gaussian_blur_sigma)
        v *= valid_mask
        v = device_obj.sim._add_noise(
            v, noise=params["noise"], noise_multiplier=params["noise_multiplier"]
        )
        v = device_obj.sim._add_hot_pixel(
            v,
            coords=params["hot_pixel_coords"],
            hot_pixel_types=params["hot_pixel_types"],
            values=params["hot_pixel_values"],
        )
        return v

    def _blur_image(self, image: np.ndarray, sigma: float = 1) -> np.ndarray:
        """Blur the image with a gaussian filter.

        Args:
            image (np.ndarray): Image to be blurred.
            sigma (float): Sigma for the gaussian filter.

        Returns:
            np.ndarray: Blurred image.
        """
        return gaussian_filter(image, sigma=sigma)

    def _create_mask(
        self,
        device_pos: np.ndarray,
        ref_motors: list[str],
        width: list[float],
        direction: list[int],
    ):
        mask = np.ones_like(device_pos)
        for ii, motor_name in enumerate(ref_motors):
            motor_pos = self.device_manager.devices.get(motor_name).obj.read()[motor_name]["value"]
            edges = [motor_pos + width[ii] / 2, motor_pos - width[ii] / 2]
            mask[..., direction[ii]] = np.logical_and(
                device_pos[..., direction[ii]] > np.min(edges),
                device_pos[..., direction[ii]] < np.max(edges),
            )

        return np.prod(mask, axis=2)



# skip-file
import socket
from unittest import mock

import numpy as np
import pytest
from ophyd import Staged
from pandablocks.connections import NeedMoreDataError

from ophyd_devices import StatusBase
from ophyd_devices.devices.panda_box.panda_box import (
    EndData,
    FrameData,
    PandaBox,
    PandaBoxDataConnection,
    PandaState,
    ReadyData,
    StartData,
)
from ophyd_devices.devices.panda_box.utils import (
    PANDA_AVAIL_PCAP_BLOCKS,
    PANDA_AVAIL_PCAP_CAPTURE_FIELDS,
    block_name_mapping,
    get_pcap_capture_fields,
)


@pytest.fixture
def _signal_aliases():
    return {"FMC_IN.VAL1.Value": "my_signal_1", "FMC_IN.VAL2.Mean": "my_signal_2"}


@pytest.fixture
def panda_box(_signal_aliases):
    return PandaBox(name="panda_box", host="localhost", signal_alias=_signal_aliases)


def test_panda_box_init_invalid_signal_alias():
    """Test that providing invalid signal aliases raises an error."""
    with pytest.raises(
        ValueError, match="Invalid signal name in signal_alias: 'Invalid.Signal.Name'"
    ):
        PandaBox(
            name="panda_box",
            host="localhost",
            signal_alias={"FMC_IN.VAL1.Value": "Invalid.Signal.Name"},
        )


def test_panda_box_init(panda_box, _signal_aliases):
    """Test initialization of PandaBox, including default signal aliases."""
    assert panda_box.name == "panda_box"
    assert panda_box.host == "localhost"
    all_signal_names = [name for name, _ in panda_box.data.signals]
    for block in PANDA_AVAIL_PCAP_BLOCKS:
        for field in PANDA_AVAIL_PCAP_CAPTURE_FIELDS:
            signal_name = f"{block}.{field}"
            if signal_name in ("FMC_IN.VAL1.Value", "FMC_IN.VAL2.Mean"):
                # These signals should be renamed
                assert _signal_aliases[signal_name] in all_signal_names
                continue
            assert (
                block_name_mapping(signal_name) in all_signal_names
            ), f"Missing signal: {signal_name}"


def test_panda_wait_for_connection(panda_box):
    """Test that wait_for_connection can be called without error."""
    with mock.patch.object(panda_box, "send_raw") as mock_send_raw:
        mock_send_raw.return_value = "OK"
        panda_box.wait_for_connection(timeout=1)
        mock_send_raw.assert_called_with("*IDN?")


def test_panda_on_connected(panda_box):
    """Test that on_connected sets the connected flag."""
    with mock.patch.object(panda_box, "data_thread") as mock_data_thread:
        mock_data_thread.is_alive.return_value = False
        panda_box.on_connected()
        mock_data_thread.start.assert_called_once()
        assert len(panda_box._data_callbacks) == 1
        cb_id = list(panda_box._data_callbacks.keys())[0]
        assert panda_box._data_callbacks[cb_id]["callback"] == panda_box._receive_frame_data
        assert panda_box._data_callbacks[cb_id]["data_type"] == PandaState.FRAME

        # Remove callback
        panda_box.remove_data_callback(cb_id)
        assert len(panda_box._data_callbacks) == 0, "Data callback was not removed"

        # Call on_connected again, should add the callback again
        mock_data_thread.reset_mock()
        mock_data_thread.is_alive.return_value = True
        panda_box.on_connected()
        mock_data_thread.start.assert_not_called()


def test_panda_add_status_callback(panda_box):
    """Test that add_status_callback adds proper status callbacks, and resolves them correctly."""

    assert panda_box.panda_state == PandaState.DISARMED, "Initial PandaBox state should be DISARMED"
    status = StatusBase(obj=panda_box)
    # I. Resolve immediately, should be successful
    panda_box.add_status_callback(status=status, success=PandaState.DISARMED, failure=[])
    status.wait(timeout=1)
    assert status.done, "Status should be done"
    assert status.success, "Status should be successful"
    assert len(panda_box._status_callbacks) == 0, "Status callback should never be added"

    # II. Resolve immediately, but with failure state, should be unsuccessful
    status = StatusBase(obj=panda_box)
    panda_box.add_status_callback(status=status, success=[], failure=PandaState.DISARMED)
    with pytest.raises(RuntimeError):
        status.wait(timeout=1)
    assert status.done, "Status should be done"
    assert not status.success, "Status should be unsuccessful"
    assert len(panda_box._status_callbacks) == 0, "Status callback should never be added"

    # III. Resolve status in success
    status = StatusBase(obj=panda_box)
    panda_box.add_status_callback(status=status, success=PandaState.READY, failure=[])
    assert len(panda_box._status_callbacks) == 1, "Status callback should be added"
    panda_box._run_status_callbacks(PandaState.START)
    assert not status.done, "Status should not be done"
    assert not status.success, "Status should not be successful"
    panda_box._run_status_callbacks(PandaState.READY)
    status.wait(timeout=1)
    assert status.done, "Status should be done"
    assert status.success, "Status should be successful"

    # IV. Resolve status in failure
    status = StatusBase(obj=panda_box)
    panda_box.add_status_callback(status=status, success=[PandaState.END], failure=PandaState.START)
    panda_box._run_status_callbacks(PandaState.START)
    with pytest.raises(RuntimeError):
        status.wait(timeout=1)
    assert status.done, "Status should be done"
    assert not status.success, "Status should be unsuccessful"


def test_panda_receive_frame_data(panda_box, _signal_aliases):
    """Test that _receive_frame_data processes data and updates signals."""
    # Create a mock FrameData
    data = np.array(
        [
            (np.float64(0), np.float64(10)),
            (np.float64(1), np.float64(11)),
            (np.float64(2), np.float64(12)),
        ],
        dtype=[("FMC_IN.VAL1.Value", "<f8"), ("COUNTER2.OUT.Value", "<f8")],
    )
    fdata = FrameData(data)
    # Use on_connected to set up data callback
    with mock.patch.object(panda_box, "data_thread") as mock_data_thread:
        mock_data_thread.is_alive.return_value = False
        panda_box.on_connected()  # This will set up the data callback
        mock_data_thread.start.assert_called_once()

        # Run the callback
        panda_box._run_data_callbacks(data=fdata, event_type=PandaState.FRAME)

        # Check that the correct signals were updated
        expected_data = {
            f"{panda_box.data.name}_my_signal_1": {
                "value": [np.float64(0), np.float64(1), np.float64(2)],
                "timestamp": mock.ANY,
            },
            f"{panda_box.data.name}_COUNTER2_OUT_Value": {
                "value": [np.float64(10), np.float64(11), np.float64(12)],
                "timestamp": mock.ANY,
            },
        }
        md = {
            "async_update": {"type": "add", "max_shape": [None]},
            "acquisition_group": panda_box._acquisition_group,
        }
        d = panda_box.data.read()
        assert d[panda_box.data.name]["value"].metadata == md, "Metadata mismatch"
        for key, v in expected_data.items():
            assert key in d[panda_box.data.name]["value"].signals, f"Missing signal: {key}"
            assert np.isclose(
                d[panda_box.data.name]["value"].signals[key]["value"], v["value"]
            ).all(), f"Incorrect values for {key}"
            assert (
                "timestamp" in d[panda_box.data.name]["value"].signals[key]
            ), f"Missing timestamp for {key}"

        assert (
            len(panda_box._data_callbacks) == 1
        ), "Data callback should still be registered after receiving data"


def test_panda_on_stop(panda_box):
    """Test that on_stop clears the data callbacks."""
    panda_box.panda_state = PandaState.READY
    panda_box.data_thread_run_event.set()  # Simulate that the data thread is running
    with mock.patch.object(panda_box, "_disarm") as mock_disarm:
        panda_box.stop()
        mock_disarm.assert_called_once()

    assert (
        panda_box.panda_state == PandaState.DISARMED
    ), "PandaBox state should be DISARMED after stop"
    # Should go back to DISARMED state
    assert (
        not panda_box.data_thread_run_event.is_set()
    ), "Data thread run event should be unset after stop"


def test_panda_on_destroy(panda_box):
    """Test that on_destroy clears the data callbacks."""
    panda_box.destroy()
    assert panda_box.data_thread_kill_event.is_set(), "Data thread kill event not set"
    assert panda_box.data_thread_run_event.is_set(), "Data thread run event not set"


def test_panda_on_stage_on_unstage(panda_box):
    """Test that on_stage sets the acquisition group."""
    panda_box.panda_state = PandaState.DISARMED
    panda_box.stage()
    assert panda_box.data_thread_run_event.is_set(), "Data thread run event not set"
    assert panda_box.staged == Staged.yes

    # Now we will unstage the panda box
    with mock.patch.object(panda_box, "_disarm") as mock_disarm:
        panda_box.unstage()
        mock_disarm.assert_called_once()
        assert panda_box.staged == Staged.no
        assert (
            not panda_box.data_thread_run_event.is_set()
        ), "Data thread run event should be unset after unstage"

    # We call on_stage again, but this time the PandaBox state is incorrect
    panda_box.panda_state = PandaState.FRAME
    with pytest.raises(RuntimeError):
        panda_box._stage_timeout_in_s = 0.1  # Set a short timeout for the test
        panda_box.stage()


def test_panda_get_signal_names_allowed_for_capture(panda_box):
    """Test that get_signal_names_allowed_for_capture returns the correct signal names."""
    return_capture = [
        "!INENC1.VAL",
        "!INENC2.VAL",
        "!INENC3.VAL",
        "!INENC4.VAL",
        "!PCAP.TS_START",
        "!PCAP.TS_END",
        "!PCAP.TS_TRIG",
        "!PCAP.GATE_DURATION",
        "!PCAP.BITS0",
        "!PCAP.BITS1",
        "!PCAP.BITS2",
        "!PCAP.BITS3",
        "!CALC1.OUT",
        "!CALC2.OUT",
        "!COUNTER1.OUT",
        "!COUNTER2.OUT",
        "!COUNTER3.OUT",
        "!COUNTER4.OUT",
        "!COUNTER5.OUT",
        "!COUNTER6.OUT",
        "!COUNTER7.OUT",
        "!COUNTER8.OUT",
        "!FILTER1.OUT",
        "!FILTER2.OUT",
        "!PGEN1.OUT",
        "!PGEN2.OUT",
        "!FMC_IN.VAL1",
        "!FMC_IN.VAL2",
        "!FMC_IN.VAL3",
        "!FMC_IN.VAL4",
        "!FMC_IN.VAL5",
        "!FMC_IN.VAL6",
        "!FMC_IN.VAL7",
        "!FMC_IN.VAL8",
        "!SFP3_SYNC_IN.POS1",
        "!SFP3_SYNC_IN.POS2",
        "!SFP3_SYNC_IN.POS3",
        "!SFP3_SYNC_IN.POS4",
        ".",
    ]

    with mock.patch.object(panda_box, "send_raw") as mock_send_raw:
        mock_send_raw.return_value = return_capture
        list_of_signals = panda_box._get_signal_names_allowed_for_capture()
        mock_send_raw.assert_called_once_with("*CAPTURE.*?")
        assert list_of_signals == PANDA_AVAIL_PCAP_BLOCKS, "Signal names mismatch"


def test_panda_get_signal_names_configured_for_capture(panda_box):
    """Test that get_signal_names_configured_for_capture returns the correct signal names."""
    return_capture = [
        "!INENC1.VAL Min Max Mean",
        "!INENC2.VAL Value",
        "!INENC3.VAL Diff",
        "!INENC4.VAL Sum",
        "!PCAP.TS_START Min Max",
    ]
    possible_signal_names = get_pcap_capture_fields()

    with mock.patch.object(panda_box, "send_raw") as mock_send_raw:
        mock_send_raw.return_value = return_capture
        list_of_signals = panda_box._get_signal_names_configured_for_capture()
        list_of_signals = [block_name_mapping(name) for name in list_of_signals]
        mock_send_raw.assert_called_once_with("*CAPTURE?")
        for signal in list_of_signals:
            assert signal in possible_signal_names, f"Unexpected signal: {signal}"


def test_panda_pre_scan_status_callback(panda_box):
    """Test that pre_scan_status_callback sets the acquisition group."""
    with mock.patch.object(panda_box, "_arm") as mock_arm:
        # I. Called with status that is not done, should do nothing
        status = StatusBase(obj=panda_box)
        panda_box._pre_scan_status_callback(status)
        mock_arm.assert_not_called()  # _arm should not be called in pre_scan_status_callback
        status.set_finished()
        # II. Called with status that is done, should arm the PandaBox
        panda_box._pre_scan_status_callback(status)
        mock_arm.assert_called_once()
        status = StatusBase(obj=panda_box)
        status.set_exception(RuntimeError("Test error"))
        # III. Called with status that is done but not successful, should do nothing
        mock_arm.reset_mock()
        panda_box._pre_scan_status_callback(status)
        mock_arm.assert_not_called()


def test_panda_get_pcap_capture_fields():
    """Test that get_pcap_capture_fields returns the correct list of capture fields."""
    expected_fields = []
    for block in PANDA_AVAIL_PCAP_BLOCKS:
        for field in PANDA_AVAIL_PCAP_CAPTURE_FIELDS:
            expected_fields.append(block_name_mapping(f"{block}.{field}"))
    actual_fields = get_pcap_capture_fields()
    assert actual_fields == expected_fields, "PCAP capture fields mismatch"


def test_panda_on_pre_scan(panda_box):
    """Test that on_pre_scan adds the correct pre-scan status callback."""
    # I. Resolve immediately in success
    panda_box.panda_state = PandaState.READY

    status = panda_box.on_pre_scan()
    assert status.done, "Status should be done"
    assert status.success, "Status should be successful"

    # II. Resolve immediately in failure
    panda_box.panda_state = PandaState.FRAME
    status = panda_box.on_pre_scan()
    with pytest.raises(RuntimeError):
        status.wait(timeout=1)
    assert status.done, "Status should be done"
    assert not status.success, "Status should be unsuccessful"

    # III. Resolve in callback
    panda_box.panda_state = PandaState.DISARMED
    status = panda_box.on_pre_scan()
    assert not status.done, "Status should not be done"
    assert not status.success, "Status should not be successful"
    panda_box._run_status_callbacks(PandaState.READY)
    status.wait(timeout=1)
    assert status.done, "Status should be done"
    assert status.success, "Status should be successful"


def test_data_loop():
    """Test that the data loop correctly handles receiving data and exiting on EndData."""
    # Mock the exact class lookup used inside PandaBox._run_data_readout
    with mock.patch(
        "ophyd_devices.devices.panda_box.panda_box.PandaBoxDataConnection"
    ) as mock_data_connection_cls:
        panda_box = PandaBox(
            name="panda_box", host="localhost", signal_alias={"FMC_IN.VAL1.Value": "my_signal_1"}
        )

        mock_data_connection = mock.MagicMock()
        mock_socket_connection = mock.MagicMock()
        mock_data_connection_cls.return_value.__enter__.return_value = (
            mock_socket_connection,
            mock_data_connection,
        )

        recv_items = [b"chunk1", b"chunk2", socket.timeout(), b"chunk3", b"chunk4", b"chunk5"]

        def recv_side_effect(_buffer_size):
            item = recv_items.pop(0)
            if isinstance(item, Exception):
                raise item
            return item

        def receive_bytes_side_effect(chunk):
            if chunk == b"chunk1":
                return [
                    StartData(
                        fields=["FMC_IN.VAL1.Value", "COUNTER2.OUT.Value"],
                        missed=0,
                        process="Scaled",
                        format="Framed",
                        sample_bytes=16,
                        arm_time=None,
                        start_time=None,
                        hw_time_offset_ns=None,
                    )
                ]
            if chunk == b"chunk2":
                return [ReadyData()]
            if chunk == b"chunk3":
                return [
                    FrameData(
                        np.array(
                            [(3, 4)],
                            dtype=[("FMC_IN.VAL1.Value", "<f8"), ("COUNTER2.OUT.Value", "<f8")],
                        )
                    )
                ]
            if chunk == b"chunk4":
                raise NeedMoreDataError()
            if chunk == b"chunk5":
                return [EndData(samples=200, reason="Ok")]
            return []

        mock_socket_connection.recv.side_effect = recv_side_effect
        mock_data_connection.receive_bytes.side_effect = receive_bytes_side_effect

        status_start = StatusBase(obj=panda_box)
        status_ready = StatusBase(obj=panda_box)
        status_frame = StatusBase(obj=panda_box)
        status_end = StatusBase(obj=panda_box)
        status_disarmed = StatusBase(obj=panda_box)

        for status, expected_state in [
            (status_start, PandaState.START),
            (status_ready, PandaState.READY),
            (status_frame, PandaState.FRAME),
            (status_end, PandaState.END),
            (status_disarmed, PandaState.DISARMED),
        ]:
            panda_box.add_status_callback(
                status=status, success=[expected_state], failure=[], check_directly=False
            )

        assert panda_box.panda_state == PandaState.DISARMED, "Initial state should be DISARMED"

        with (
            mock.patch.object(panda_box.data, "put"),
            mock.patch.object(panda_box, "_disarm") as mock_disarm,
        ):
            panda_box.data_thread_run_event.set()
            panda_box._run_data_readout()

            for status in [status_start, status_ready, status_frame, status_end, status_disarmed]:
                assert status.done, f"{status} should be done"
                assert status.success, f"{status} should be successful"

        mock_data_connection_cls.assert_called_once_with(
            "localhost", scaled=True, socket_timeout=0.1
        )
        mock_disarm.assert_called_once(), "_disarm should be called once after receiving EndData"

        assert (
            not panda_box.data_thread_run_event.is_set()
        ), "Data thread run event should be unset after receiving EndData"


def test_pandabox_data_connection():
    """Test that PandaBoxDataConnection correctly connects and receives data."""
    with mock.patch("socket.socket") as mock_socket_cls:
        mock_socket = mock.MagicMock()
        mock_socket_cls.return_value = mock_socket

        with PandaBoxDataConnection("localhost", scaled=True, socket_timeout=0.1) as (
            data_socket,
            data_connection,
        ):
            mock_socket.connect.assert_called_once_with(("localhost", 8889))
            expected_connect_command = "XML FRAMED SCALED\n"  # Connect command for scaled data
            mock_socket.sendall.assert_called_once_with(expected_connect_command.encode())

        # Simulate exiting the context manager, which should close the socket
        mock_socket.close.assert_called_once()



"""
This module provides a class for creating a pseudo signal that is computed from other signals.
"""

import ast
import re
from functools import reduce
from typing import Callable

import numpy as np
import scipy as sp
from bec_lib import bec_logger
from ophyd import Signal, SignalRO
from ophyd.ophydobj import Kind

logger = bec_logger.logger

_FIRST_DEF = re.compile(r"(^|\\n)\s*def", re.MULTILINE)


def _smart_strip(method: str) -> str:
    if (first_def := _FIRST_DEF.search(method)) is not None:
        return method[first_def.span()[1] - 3 :]
    else:
        raise ValueError(f"No 'def' keyword found in function definition: {method}")


def rgetattr(obj, attr, *args):
    """See https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-objects"""

    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return reduce(_getattr, [obj] + attr.split("."))


class ComputedSignal(SignalRO):
    """
    A read-only signal that is computed from other signals. The compute method should be a string
    representation of a function that takes the input signals as arguments and returns the computed
    value. The input signals should be provided as a list of strings that represent the path to the
    signal in the device manager.
    """

    def __init__(
        self,
        *,
        name,
        value=0,
        timestamp=None,
        device_manager=None,
        parent=None,
        labels=None,
        kind=Kind.hinted,
        tolerance=None,
        rtolerance=None,
        metadata=None,
        cl=None,
        attr_name="",
    ):
        super().__init__(
            name=name,
            value=value,
            timestamp=timestamp,
            parent=parent,
            labels=labels,
            kind=kind,
            tolerance=tolerance,
            rtolerance=rtolerance,
            metadata=metadata,
            cl=cl,
            attr_name=attr_name,
        )
        self._device_manager = device_manager
        self._input_signals: list[Signal] = []
        self._signal_subs = []
        self._compute_method = None
        self._compute_method_str = None

    def _signal_callback(self, *args, **kwargs):
        self._run_subs(sub_type=self.SUB_VALUE, old_value=None, value=self.get())

    @property
    def compute_method(self) -> Callable | None:
        """
        Property that returns the compute method for the pseudo signal.
        Example:
            >>> signal.compute_method = "def test(a, b): return a.get() + b.get()"

        """
        return self._compute_method_str

    @compute_method.setter
    def compute_method(self, method: str):
        """
        Set the compute method for the pseudo signal. We  import numpy (as np) and scipy as packages
        that are also available for user to use in their compute method.

        Args:
            compute_method (str): The compute method to be used. This should be a string
                representation of a function that takes the input signals as arguments
                and returns the computed value.

        Example:
            >>> signal.compute_method = "def test(a, b): return a.get() + b.get()"

        """
        logger.info(f"Updating compute method for {self.name}.")
        method = _smart_strip(method)
        # Parse and validate the function using AST
        try:
            tree = ast.parse(method)
            if not tree.body or not isinstance(tree.body[0], ast.FunctionDef):
                raise ValueError("The compute method should be a valid function definition")

            # Rename the function in the AST
            func_def = tree.body[0]
            func_def.name = "user_compute_method"

            # Convert AST back to code and compile
            code = compile(tree, "<string>", "exec")
        except SyntaxError as exc:
            raise ValueError(f"Invalid function syntax: {exc}") from exc

        self._compute_method_str = method
        # pylint: disable=exec-used
        namespace = {"np": np, "sp": sp}
        exec(code, namespace)
        self._compute_method = namespace["user_compute_method"]

    @property
    def input_signals(self):
        """
        Set the input signals for the pseudo signal

        Args:
            *input_vars: The input signals to be used for the computation

        Example:
            >>> signal.input_signals = ["samx.readback", "samx.readback"]

        """
        return self._input_signals

    @input_signals.setter
    def input_signals(self, input_vars):
        if self._signal_subs:
            for signal, sub_id in self._signal_subs:
                signal.unsubscribe(sub_id)
        signals = []
        for signal in input_vars:
            if isinstance(signal, str):
                obj = rgetattr(self._device_manager.devices, signal)
                sub_id = obj.subscribe(self._signal_callback)
                self._signal_subs.append((obj, sub_id))
                signals.append(obj)
            else:
                signals.append(signal)
        self._input_signals = signals

    def get(self):
        if self._compute_method:
            # pylint: disable=not-callable
            if self.input_signals:
                return self._compute_method(*self.input_signals)
            return self._compute_method()
        return None



from ophyd import SoftPositioner as _SoftPositioner


class SoftPositioner(_SoftPositioner):
    """
    A patched version of ophyd's SoftPositioner that complies with
    ophyd device protocol.
    """

    def __init__(self, *, egu="", limits=None, source="computed", init_pos=None, **kwargs):
        super().__init__(egu=egu, limits=limits, source=source, init_pos=init_pos, **kwargs)
        self._destroyed = False

    def destroy(self):
        self._destroyed = True
        super().destroy()



import h5py

# Necessary import to allow h5py to open compressed h5files.
# pylint: disable=unused-import
import hdf5plugin  # noqa: F401
import numpy as np
from ophyd import Component, Kind, Staged
from scipy.ndimage import gaussian_filter

from ophyd_devices.sim.sim_frameworks.device_proxy import DeviceProxy


class H5ImageReplayProxy(DeviceProxy):
    """This Proxy class can be used to replay images from an h5 file.

    If the number of requested images is larger than the number of available images,
    the images will be replayed from the beginning.

    An example for the configuration of this is device is in ophyd_devices.configs.ophyd_devices_simulation.yaml
    """

    USER_ACCESS = ["file_source", "h5_entry"]

    def __init__(self, name, *args, device_manager=None, **kwargs):
        self.h5_file = None
        self.h5_dataset = None
        self._number_of_images = None
        self._staged = Staged.no
        self._image = None
        self._index = 0
        self._file_source = ""
        self._h5_entry = ""
        super().__init__(name, *args, device_manager=device_manager, **kwargs)

    @property
    def file_source(self) -> str:
        """File source property."""
        return self._file_source

    @file_source.setter
    def file_source(self, file_source: str) -> None:
        self.config[list(self.config.keys())[0]]["file_source"] = file_source
        self._file_source = file_source

    @property
    def h5_entry(self) -> str:
        """H5 entry property."""
        return self._h5_entry

    @h5_entry.setter
    def h5_entry(self, h5_entry: str) -> None:
        self.config[list(self.config.keys())[0]]["h5_entry"] = h5_entry
        self._h5_entry = h5_entry

    def _update_device_config(self, config: dict) -> None:
        super()._update_device_config(config)
        if len(config.keys()) > 1:
            raise RuntimeError(
                f"The current implementation of device {self.name} can only replay data for a single device. The config has information about multiple devices {config.keys()}"
            )
        self._init_signals()

    def _init_signals(self):
        """Initialize the signals for the device."""
        if "file_source" in self.config[list(self.config.keys())[0]]:
            self.file_source = self.config[list(self.config.keys())[0]]["file_source"]
        if "h5_entry" in self.config[list(self.config.keys())[0]]:
            self.h5_entry = self.config[list(self.config.keys())[0]]["h5_entry"]

    def _open_h5_file(self) -> None:
        """Opens the HDF5 file found in the file_source signal and the HDF5 dataset specified by the h5_entry signal."""
        self.h5_file = h5py.File(self.file_source, mode="r")
        self.h5_dataset = self.h5_file[self.h5_entry]
        self._number_of_images = self.h5_dataset.shape[0]

    def _close_h5_file(self) -> None:
        """Close the HDF5 file."""
        self.h5_file.close()

    def stop(self) -> None:
        """Stop the device."""
        if self.h5_file:
            self._close_h5_file()
        self.h5_file = None
        self.h5_dataset = None
        self._number_of_images = None
        self._index = 0

    def stage(self) -> list[object]:
        """Stage the device.
        This opens the HDF5 dataset, unstaging will close it.
        """

        if self._staged != Staged.no:
            return [self]
        try:
            self._open_h5_file()
        except Exception as exc:
            if self.h5_file:
                self.stop()
            raise FileNotFoundError(
                f"Could not open h5file {self.file_source} or access data set {self.h5_dataset} in file"
            ) from exc

        self._staged = Staged.yes
        return [self]

    def unstage(self) -> list[object]:
        """Unstage the device, also closes the HDF5 dataset"""
        if self.h5_file:
            self.stop()
        self._staged = Staged.no
        return [self]

    def _load_image(self):
        """Try loading the image from the h5 dataset, and set it to self._image."""
        if self.h5_file:
            slice_nr = self._index % self._number_of_images
            self._index = self._index + 1
            self._image = self.h5_dataset[slice_nr, ...]
            return
        try:
            self.stage()
            slice_nr = self._index % self._number_of_images
            self._index = self._index + 1
            self._image = self.h5_dataset[slice_nr, ...]
            self.unstage()
        except Exception as exc:
            raise FileNotFoundError(
                f"Could not open h5file {self.file_source} or access data set {self.h5_dataset} in file"
            ) from exc

    def _compute(self, device_name: str, *args, **kwargs) -> np.ndarray:
        """Compute the image.

        Returns:
            np.ndarray: Image.
        """
        self._load_image()
        return self._image



"""Simulated 2D camera device"""

import numpy as np
from bec_lib.logger import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, Kind, StatusBase

from ophyd_devices.interfaces.base_classes.psi_device_base import PSIDeviceBase
from ophyd_devices.sim.sim_data import SimulatedDataCamera
from ophyd_devices.sim.sim_signals import ReadOnlySignal, SetableSignal
from ophyd_devices.sim.sim_utils import H5Writer
from ophyd_devices.utils.bec_signals import PreviewSignal

logger = bec_logger.logger


class SimCameraControl(Device):
    """SimCamera Control layer"""

    USER_ACCESS = ["sim", "registered_proxies"]

    sim_cls = SimulatedDataCamera
    SHAPE = (100, 100)
    BIT_DEPTH = np.uint16

    SUB_MONITOR = "device_monitor_2d"
    _default_sub = SUB_MONITOR

    exp_time = Cpt(SetableSignal, name="exp_time", value=1, kind=Kind.config)
    file_pattern = Cpt(SetableSignal, name="file_pattern", value="", kind=Kind.config)
    frames = Cpt(SetableSignal, name="frames", value=1, kind=Kind.config)
    burst = Cpt(SetableSignal, name="burst", value=1, kind=Kind.config)

    image_shape = Cpt(SetableSignal, name="image_shape", value=SHAPE, kind=Kind.config)
    image = Cpt(
        ReadOnlySignal,
        name="image",
        value=np.empty(SHAPE, dtype=BIT_DEPTH),
        compute_readback=True,
        kind=Kind.omitted,
    )
    preview = Cpt(PreviewSignal, name="preview", ndim=2, num_rotation_90=0)
    write_to_disk = Cpt(SetableSignal, name="write_to_disk", value=False, kind=Kind.config)

    def __init__(self, name, *, parent=None, sim_init: dict = None, device_manager=None, **kwargs):
        self.sim_init = sim_init
        self.device_manager = device_manager
        self._registered_proxies = {}
        self.sim = self.sim_cls(parent=self, **kwargs)
        self.h5_writer = H5Writer()
        super().__init__(name=name, parent=parent, **kwargs)
        if self.sim_init:
            self.sim.set_init(self.sim_init)

    @property
    def registered_proxies(self) -> None:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies


class SimCamera(PSIDeviceBase, SimCameraControl):
    """A simulated device mimic any 2D camera.

    It's image is a computed signal, which is configurable by the user and from the command line.
    The corresponding simulation class is sim_cls=SimulatedDataCamera, more details on defaults within the simulation class.

    >>> camera = SimCamera(name="camera")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argument, passed on to all signals of the device.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    sim_init (dict)         : Dictionary to initiate parameters of the simulation, check simulation type defaults for more details.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.

    """

    def __init__(self, name: str, scan_info=None, device_manager=None, **kwargs):
        super().__init__(name=name, scan_info=scan_info, device_manager=device_manager, **kwargs)
        self.file_path = None

    def on_trigger(self) -> StatusBase:
        """Trigger the camera to acquire images.

        This method can be called from BEC during a scan. It will acquire images and send them to BEC.
        Whether the trigger is send from BEC is determined by the softwareTrigger argument in the device config.

        Here, we also run a callback on SUB_MONITOR to send the image data the device_monitor endpoint in BEC.
        """

        def trigger_cam() -> None:
            """Trigger the camera to acquire images."""
            for _ in range(self.burst.get()):
                data = self.image.get()
                # pylint: disable=protected-access
                self.preview.put(data)
                self._run_subs(sub_type=self.SUB_MONITOR, value=data)
                if self.write_to_disk.get():
                    self.h5_writer.receive_data(data)

        status = self.task_handler.submit_task(trigger_cam)
        return status

    def on_stage(self) -> None:
        """Stage the camera for upcoming scan

        This method is called from BEC in preparation of a scan.
        It receives metadata about the scan from BEC,
        compiles it and prepares the camera for the scan.

        FYI: No data is written to disk in the simulation, but upon each trigger it
        is published to the device_monitor endpoint in REDIS.
        """
        self.file_path = self.file_utils.get_full_path(
            scan_status_msg=self.scan_info.msg, name=self.name
        )
        self.frames.set(
            self.scan_info.msg.num_points * self.scan_info.msg.scan_parameters["frames_per_trigger"]
        ).wait()
        self.exp_time.set(self.scan_info.msg.scan_parameters["exp_time"]).wait()
        self.burst.set(self.scan_info.msg.scan_parameters["frames_per_trigger"]).wait()
        if self.write_to_disk.get():
            self.h5_writer.on_stage(file_path=self.file_path, h5_entry="/entry/data/data")
            # pylint: disable=protected-access
            self._run_subs(
                sub_type=self.SUB_FILE_EVENT,
                file_path=self.file_path,
                done=False,
                successful=False,
                hinted_location={"data": "/entry/data/data"},
            )

    def on_complete(self) -> StatusBase:
        """Complete the motion of the simulated device."""

        if not self.write_to_disk.get():
            return None

        def complete_cam():
            """Complete the camera acquisition."""
            self.h5_writer.on_complete()
            self._run_subs(
                sub_type=self.SUB_FILE_EVENT,
                file_path=self.file_path,
                done=True,
                successful=True,
                hinted_location={"data": "/entry/data/data"},
            )

        status = self.task_handler.submit_task(complete_cam)
        return status

    def on_unstage(self) -> None:
        """Unstage the camera device."""
        if self.write_to_disk.get():
            self.h5_writer.on_unstage()

    def on_stop(self) -> None:
        """Stop the camera acquisition."""
        self.task_handler.shutdown()
        self.on_unstage()



import threading
import time

import numpy as np
from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, DeviceStatus, Kind
from ophyd.flyers import FlyerInterface
from ophyd.status import StatusBase

from ophyd_devices.sim.sim_data import SimulatedPositioner
from ophyd_devices.sim.sim_signals import ReadOnlySignal

logger = bec_logger.logger


class SimFlyer(Device, FlyerInterface):
    """A simulated device mimicking any 2D Flyer device (position, temperature, rotation).

    The corresponding simulation class is sim_cls=SimulatedPositioner, more details on defaults within the simulation class.

    >>> flyer = SimFlyer(name="flyer")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argument, passed on to all signals of the device.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation, device_manager is passed on automatically.
    """

    USER_ACCESS = ["sim", "registered_proxies"]

    sim_cls = SimulatedPositioner

    readback = Cpt(
        ReadOnlySignal, name="readback", value=0, kind=Kind.hinted, compute_readback=False
    )

    def __init__(
        self,
        name: str,
        *,
        precision: int = 3,
        parent=None,
        kind=None,
        device_manager=None,
        sim_init: dict = None,
        # TODO remove after refactoring config
        delay: int = 1,
        update_frequency: int = 100,
        **kwargs,
    ):

        self.sim = self.sim_cls(parent=self, **kwargs)
        self.sim_init = sim_init
        self.precision = precision
        self.device_manager = device_manager
        self._registered_proxies = {}

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self.sim.sim_state[self.name] = self.sim.sim_state.pop(self.readback.name, None)
        self.readback.name = self.name
        if self.sim_init:
            self.sim.set_init(self.sim_init)

    @property
    def registered_proxies(self) -> None:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies

    @property
    def hints(self):
        """Return the hints of the simulated device."""
        return {"fields": ["flyer_samx", "flyer_samy"]}

    @property
    def egu(self) -> str:
        """Return the engineering units of the simulated device."""
        return "mm"

    def complete(self) -> StatusBase:
        """Complete the motion of the simulated device."""
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def kickoff(self, metadata, num_pos, positions, exp_time: float = 0):
        """Kickoff the flyer to execute code during the scan."""
        positions = np.asarray(positions)

        def produce_data(device, metadata):
            """Simulate the data being produced by the flyer."""
            buffer_time = 0.2
            elapsed_time = 0
            bundle = messages.BundleMessage()
            for ii in range(num_pos):
                bundle.append(
                    messages.DeviceMessage(
                        signals={
                            "flyer_samx": {"value": positions[ii, 0], "timestamp": 0},
                            "flyer_samy": {"value": positions[ii, 1], "timestamp": 0},
                        },
                        metadata={"point_id": ii, **metadata},
                    )
                )
                time.sleep(exp_time)
                elapsed_time += exp_time
                if elapsed_time > buffer_time:
                    elapsed_time = 0
                    logger.info(f"Sending data point {ii} for {device.name}.")
                    device.device_manager.connector.set_and_publish(
                        MessageEndpoints.device_read(device.name), bundle
                    )
                    bundle = messages.BundleMessage()
                    device.device_manager.connector.set(
                        MessageEndpoints.device_status(device.name),
                        messages.DeviceStatusMessage(
                            device=device.name, status=1, metadata={"point_id": ii, **metadata}
                        ),
                    )
            device.device_manager.connector.set_and_publish(
                MessageEndpoints.device_read(device.name), bundle
            )
            device.device_manager.connector.set(
                MessageEndpoints.device_status(device.name),
                messages.DeviceStatusMessage(
                    device=device.name, status=0, metadata={"point_id": num_pos, **metadata}
                ),
            )
            print("done")

        flyer = threading.Thread(target=produce_data, args=(self, metadata))
        flyer.start()



from unittest import mock

import pytest
from bec_server.device_server.tests.utils import DMMock

from ophyd_devices.utils.dynamic_pseudo import ComputedSignal, _smart_strip


@pytest.fixture
def device_manager_with_devices():
    dm = DMMock()
    dm.add_device("a")
    dm.add_device("b")
    device_mock = mock.MagicMock()
    dm.devices["a"] = device_mock
    dm.devices["b"] = device_mock

    return dm


@pytest.mark.parametrize(
    "compute_method_str",
    [
        "def test(a, b): return a.get() + b.get()",
        "def   test(a,   b): return a.get() + b.get()",
        "    def my_compute_method(a,b):\n        return a.get() + b.get()\n",
        "#comment goes here\n    def my_compute_method(a,b):\n        return a.get() + b.get()\n",
        "#comment goes here\n\n\n    def my_compute_method(a,b):\n        return a.get() + b.get()\n",
        "#comment goes here\n\n\n    def my_compute_method(a,b):\n#comment inside\n        return a.get() + b.get()\n",
    ],
)
def test_computed_signal(device_manager_with_devices, compute_method_str):
    signal = ComputedSignal(name="test", device_manager=device_manager_with_devices)
    assert signal.get() is None

    # Configure the mocks before setting input signals
    device_manager_with_devices.devices["a"].readback.get.return_value = 20
    device_manager_with_devices.devices["b"].readback.get.return_value = 20

    signal.compute_method = compute_method_str
    signal.input_signals = ["a.readback", "b.readback"]

    assert signal.get() == 40

    # pylint: disable=protected-access
    assert callable(signal._compute_method)
    assert signal._compute_method_str == _smart_strip(compute_method_str)



from abc import ABC
from typing import TypedDict

from ophyd import Component as Cpt
from ophyd import Device
from ophyd.device import required_for_connection
from ophyd.positioner import PositionerBase
from ophyd.signal import EpicsSignalBase, Signal
from ophyd.status import MoveStatus
from ophyd.status import wait as status_wait
from ophyd.utils.epics_pvs import AlarmSeverity, fmt_time


class _SignalSentinel(object): ...


_OPTIONAL_SIGNAL = _SignalSentinel()
_REQUIRED_SIGNAL = _SignalSentinel()
_SIGNAL_NOT_AVAILABLE = "Signal not available"


class PSIPositionerException(Exception): ...


class RequiredSignalNotSpecified(PSIPositionerException): ...


class OptionalSignalNotSpecified(PSIPositionerException): ...


class SimplePositionerSignals(TypedDict, total=False):
    """The list of all the signals in the PSISimplePositionerBase"""

    user_readback: str
    user_setpoint: str
    motor_done_move: str
    velocity: str
    motor_stop: str


_SIMPLE_SIGNAL_NAMES = SimplePositionerSignals.__optional_keys__


class PositionerSignals(SimplePositionerSignals, total=False):
    """The list of all the signals in the PSIPositionerBase. See that class for
    documentation of signal functionality."""

    user_offset: str
    user_offset_dir: str
    offset_freeze_switch: str
    set_use_switch: str
    acceleration: str
    motor_egu: str
    motor_is_moving: str
    high_limit_switch: str
    low_limit_switch: str
    high_limit_travel: str
    low_limit_travel: str
    direction_of_travel: str
    home_forward: str
    home_reverse: str
    tolerated_alarm: str


_SIGNAL_NAMES = PositionerSignals.__optional_keys__


class PSISimplePositionerBase(ABC, Device, PositionerBase):
    """Base class for simple positioners."""

    SIGNAL_NAMES = _SIMPLE_SIGNAL_NAMES

    user_readback: EpicsSignalBase = _REQUIRED_SIGNAL
    user_setpoint: EpicsSignalBase = _OPTIONAL_SIGNAL
    velocity: EpicsSignalBase = _OPTIONAL_SIGNAL
    motor_stop: EpicsSignalBase = _OPTIONAL_SIGNAL
    motor_done_move: EpicsSignalBase = _OPTIONAL_SIGNAL

    stop_value = 1  # The value to put to the stop PV (if set) to make the motor stop
    done_value = 1  # The value expected to be reported by motor_done_move when the move is done
    use_put_complete = False  # Whether to use put-completion for the setpoint for the move status

    def __init__(
        self,
        prefix="",
        *,
        name,
        limits: list[float] | tuple[float, ...] | None = None,
        deadband: float | None = None,
        use_put_completion: bool | None = None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        override_suffixes: SimplePositionerSignals = {},
        **kwargs,
    ):
        """A simple positioner class, to provide the functionality of PVPositioner but behave more
        similarly to EpicsMotor.

        Args:
            name (str): (required) the name of the device
            limits (list | tuple | None): If given, a length-2 sequence within the range of which movement is allowed.
            deadband (float | None): If given, set a soft deadband of this absolute value, within which positioner moves will return immediately. If the positioner has no motor_done_move signal, you must provide this.
            use_put_completion (bool | None): If given, use put completion on the setpoint signal to resolve the move status.
            override_suffixes (dict[str, str]): a dictionary of signal_name: pv_suffix which will replace the values in the signal classvar.
        """
        super().__init__(
            prefix=prefix,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            name=name,
            parent=parent,
            **kwargs,
        )

        if limits is not None:
            self._limits = tuple(limits)
        else:
            self._limits = None
        self._deadband = deadband
        if use_put_completion is not None:
            self.use_put_complete = use_put_completion
        self._egu = kwargs.get("egu") or ""

        if (missing := self._remaining_defaults(_REQUIRED_SIGNAL)) != set():
            raise RequiredSignalNotSpecified(f"Signal(s) {missing} must be defined in a subclass")

        if self.user_readback is _OPTIONAL_SIGNAL and self.motor_done_move is _OPTIONAL_SIGNAL:
            raise ValueError(
                "Positioner must have at least one of user_readback and motor_done_move"
            )

        not_implemented = self._remaining_defaults(_OPTIONAL_SIGNAL)
        for signal_name, pv_suffix in override_suffixes.items():
            if signal_name not in not_implemented:
                component: Cpt[EpicsSignalBase] = getattr(self.__class__, signal_name)
                signal: EpicsSignalBase = getattr(self, signal_name)
                signal.__init__(
                    prefix + pv_suffix,
                    name=self.name + self._child_name_separator + signal_name,
                    **component.kwargs,
                )
            else:
                self.log.warning(
                    f"{self.__class__} does not implement overridden signal {signal_name}"
                )

        if self.user_readback is not _OPTIONAL_SIGNAL:
            self.user_readback.subscribe(self._pos_changed)
        if self.motor_done_move is not _OPTIONAL_SIGNAL:
            self.motor_done_move.subscribe(self._move_changed)
        elif deadband is None:
            raise ValueError("Deadband must not be None for a device with no done signal")

        # Make the default alias for the user_readback the name of the motor itself
        # for compatibility with EpicsMotor
        self.user_readback.name = self.name

    def _remaining_defaults(self, attr: _SignalSentinel) -> set[str]:
        return set(filter(lambda s: getattr(self, s) is attr, self.SIGNAL_NAMES))

    @property
    def egu(self):
        """The engineering units (EGU) for a position"""
        return self._egu

    def check_value(self, pos):
        """Check that the position is within the soft limits"""
        if self.limits is not None:
            low, high = self.limits
            if low != high and not (low <= pos <= high):
                raise ValueError(f"{pos} outside of user-specified limits {self.limits}")
        else:
            self.user_setpoint.check_value(pos)

    @property
    def moving(self):
        """Whether or not the motor is moving

        If a `done` PV is specified, it will be read directly to get the motion
        status. If not, it determined from the internal state of PVPositioner.

        Returns
        -------
        bool
        """
        dval = self.motor_done_move.get(use_monitor=False)
        return dval != self.done_value

    def _setup_move(self, position):
        """Move and do not wait until motion is complete (asynchronous)"""
        self.log.debug(f"{self.name}.user_setpoint = {position}")
        if not self.use_put_complete:
            self.user_setpoint.put(position, wait=True)
        else:
            self.user_setpoint.put(
                position, wait=False, callback=lambda *_: self._done_moving(success=True)
            )

    def move(self, position, wait=True, timeout=None, moved_cb=None):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback must
            accept one keyword argument: 'obj' which will be set to this
            positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """

        if self._deadband is not None and abs(position - self._position) < self._deadband:
            return MoveStatus(self, position, done=True, success=True)

        status = super().move(position, timeout=timeout, moved_cb=moved_cb)

        try:
            self._setup_move(position)
            if wait:
                status_wait(status)
        except KeyboardInterrupt:
            self.stop()
            raise

        return status

    @required_for_connection
    def _move_changed(self, timestamp=None, value=None, sub_type=None, **kwargs):
        was_moving = self._moving
        self._moving = value != self.done_value

        started = False
        if not self._started_moving:
            started = self._started_moving = not was_moving and self._moving
            self.log.debug(f"[ts={fmt_time(timestamp)}] {self.name} started moving: {started}")

        self.log.debug(
            f"[ts={fmt_time(timestamp)}] {self.name} moving: {self._moving} (value={value})"
        )
        if started:
            self._run_subs(sub_type=self.SUB_START, timestamp=timestamp, value=value, **kwargs)

        if self.motor_done_move is _OPTIONAL_SIGNAL:
            # if there is no motor_done_move, we came here from self._pos_changed with a value
            # based on whether whe are within the deadband
            if not self._moving:
                # we got a position update within the deadband of the setpoint, close out move statuses.
                self._run_subs(
                    sub_type=self._SUB_REQ_DONE, timestamp=timestamp, value=value, **kwargs
                )

    @required_for_connection
    def _pos_changed(self, timestamp=None, value=None, **kwargs):
        """Callback from EPICS, indicating a change in position"""
        self._set_position(value)
        if self.motor_done_move is _OPTIONAL_SIGNAL:
            # No DMOV, so we compare to the setpoint to see if we are done
            self._move_changed(
                timestamp=timestamp,
                value=(abs(self.user_setpoint.get() - value) < self._deadband),
                **kwargs,
            )

    def stop(self, *, success=False):
        if self.motor_stop is not _OPTIONAL_SIGNAL:
            self.motor_stop.put(self.stop_value, wait=False)
        super().stop(success=success)

    @property
    def report(self):
        rep = super().report
        rep["pv"] = self.user_readback.pvname
        return rep

    @property
    def limits(self):
        if self._limits is not None:
            return tuple(self._limits)
        else:
            return self.user_setpoint.limits

    @property
    def egu(self):
        return self._egu

    def _repr_info(self):
        yield from super()._repr_info()

        yield ("limits", self._limits)
        yield ("egu", self._egu)

    def _done_moving(self, **kwargs):
        self._move_changed(value=self.done_value)
        super()._done_moving(**kwargs)


class PSIPositionerBase(PSISimplePositionerBase):
    """Base class for positioners which are similar to a motor but do not implement
    all the required signals for an EpicsMotor or have different PV suffices."""

    SIGNAL_NAMES = _SIGNAL_NAMES

    # calibration dial <-> user
    # https://epics.anl.gov/bcda/synApps/motor/motorRecord.html#Fields_calib
    user_offset = _OPTIONAL_SIGNAL
    user_offset_dir = _OPTIONAL_SIGNAL
    # Fix the difference between the user and dial positions
    offset_freeze_switch = _OPTIONAL_SIGNAL
    set_use_switch = _OPTIONAL_SIGNAL

    # configuration
    acceleration = _OPTIONAL_SIGNAL
    motor_egu = _OPTIONAL_SIGNAL

    # motor status
    motor_is_moving = _OPTIONAL_SIGNAL
    high_limit_switch = _OPTIONAL_SIGNAL
    low_limit_switch = _OPTIONAL_SIGNAL
    high_limit_travel: Signal | _SignalSentinel = _OPTIONAL_SIGNAL
    low_limit_travel: Signal | _SignalSentinel = _OPTIONAL_SIGNAL
    direction_of_travel = _OPTIONAL_SIGNAL

    # commands
    home_forward = _OPTIONAL_SIGNAL
    home_reverse = _OPTIONAL_SIGNAL

    # alarm information
    tolerated_alarm = AlarmSeverity.NO_ALARM

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        child_name_separator="_",
        **kwargs,
    ):
        super().__init__(
            prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            child_name_separator=child_name_separator,
            **kwargs,
        )
        if self.limits is not None:
            for sig, lim in (
                (self.low_limit_travel, self.limits[0]),
                (self.high_limit_travel, self.limits[1]),
            ):
                # If the limit signals are defined as soft signals, propagate the limits there
                if sig is not _OPTIONAL_SIGNAL and type(sig) is Signal:
                    sig.put(lim)

    @property
    def egu(self):
        if self.motor_egu is not _OPTIONAL_SIGNAL:
            return self.motor_egu.get()
        return self._egu



"""Module for a simulated 1D Waveform detector, i.e. a Falcon XRF detector."""

import os
import threading
import time
import traceback
from typing import Any

import numpy as np
from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, DeviceStatus, Kind, Staged
from typeguard import typechecked

from ophyd_devices.sim.sim_data import SimulatedDataWaveform
from ophyd_devices.sim.sim_signals import ReadOnlySignal, SetableSignal
from ophyd_devices.utils import bec_utils
from ophyd_devices.utils.bec_signals import AsyncMultiSignal, AsyncSignal, ProgressSignal
from ophyd_devices.utils.errors import DeviceStopError

logger = bec_logger.logger


class AsyncUpdateSignal(SetableSignal):
    """Async updated signal, with check for async_update type."""

    def check_value(self, value, **kwargs) -> None:
        """Check the value of the async_update signal."""
        if value not in ["add_slice", "add"]:
            raise ValueError(f"Invalid async_update type: {value} for signal {self.name}")

    # FIXME: BEC issue #443 remove this method once tests in BEC are updated.
    def put(self, value: Any) -> None:
        """Put the value of the async_update signal."""
        if value in ["append", "extend"]:
            if value == "append":
                logger.warning(
                    f"Deprecated async_update of type {value} for signal {self.name}, falling back to 'add_slice'"
                )
                value = "add_slice"
            elif value == "extend":
                logger.warning(
                    f"Deprecated async_update of type {value} for signal {self.name}, falling back to 'add'"
                )
                value = "add"
        super().put(value)


class SimWaveform(Device):
    """A simulated device mimic any 1D Waveform detector.

    It's waveform is a computed signal, which is configurable by the user and from the command line.
    The corresponding simulation class is sim_cls=SimulatedDataWaveform, more details on defaults within the simulation class.

    >>> waveform = SimWaveform(name="waveform")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argument, passed on to all signals of the device.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    sim_init (dict)         : Dictionary to initiate parameters of the simulation, check simulation type defaults for more details.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation, device_manager is passed on automatically.

    """

    USER_ACCESS = ["sim", "registered_proxies", "delay_slice_update"]

    sim_cls = SimulatedDataWaveform
    SHAPE = (1000,)
    BIT_DEPTH = np.uint16

    SUB_MONITOR = "device_monitor_1d"
    _default_sub = SUB_MONITOR

    exp_time = Cpt(SetableSignal, name="exp_time", value=1, kind=Kind.config)
    file_path = Cpt(SetableSignal, name="file_path", value="", kind=Kind.config)
    file_pattern = Cpt(SetableSignal, name="file_pattern", value="", kind=Kind.config)
    frames = Cpt(SetableSignal, name="frames", value=1, kind=Kind.config)
    burst = Cpt(SetableSignal, name="burst", value=1, kind=Kind.config)

    waveform_shape = Cpt(SetableSignal, name="waveform_shape", value=SHAPE, kind=Kind.config)
    waveform = Cpt(
        ReadOnlySignal,
        name="waveform",
        value=np.empty(SHAPE, dtype=BIT_DEPTH),
        compute_readback=True,
        kind=Kind.hinted,
    )
    waveform_0d = Cpt(AsyncSignal, name="waveform_0d", ndim=0, max_size=1000, kind=Kind.hinted)
    data = Cpt(AsyncSignal, name="data", ndim=1, max_size=1000)
    # Can be extend or append
    async_update = Cpt(AsyncUpdateSignal, value="add", kind=Kind.config)
    progress = Cpt(ProgressSignal, name="progress")
    async_multi_data = Cpt(
        AsyncMultiSignal, name="async_multi_data", signals=["data1", "data2"], ndim=1, max_size=1000
    )
    slice_size = Cpt(SetableSignal, value=100, dtype=np.int32, kind=Kind.config)

    def __init__(
        self,
        name,
        *,
        kind=None,
        parent=None,
        sim_init: dict = None,
        device_manager=None,
        scan_info=None,
        **kwargs,
    ):
        self.sim_init = sim_init
        self._registered_proxies = {}
        self.sim = self.sim_cls(parent=self, **kwargs)

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        if device_manager:
            self.device_manager = device_manager
        else:
            self.device_manager = bec_utils.DMMock()

        self.connector = self.device_manager.connector
        self._stream_ttl = 1800  # 30 min max
        self.stopped = False
        self._staged = Staged.no
        self._trigger_thread = None
        self._trigger_received = 0
        self.scan_info = scan_info
        self._delay_slice_update = False
        if self.sim_init:
            self.sim.set_init(self.sim_init)
        self._slice_index = 0

    @property
    def delay_slice_update(self) -> bool:
        """Delay updates in-between slices specified by waveform_shape and slice_size."""
        return self._delay_slice_update

    @delay_slice_update.setter
    @typechecked
    def delay_slice_update(self, value: bool) -> None:
        self._delay_slice_update = value

    @property
    def registered_proxies(self) -> dict[str, Any]:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies

    def trigger(self) -> DeviceStatus:
        """Trigger the camera to acquire images.

        This method can be called from BEC during a scan. It will acquire images and send them to BEC.
        Whether the trigger is send from BEC is determined by the softwareTrigger argument in the device config.

        Here, we also run a callback on SUB_MONITOR to send the image data the device_monitor endpoint in BEC.
        """
        status = DeviceStatus(self)
        self.waveform_0d.put(
            np.random.randint(0, 100), async_update={"type": "add", "max_shape": [None]}
        )

        def acquire(status: DeviceStatus):
            try:
                for _ in range(self.burst.get()):
                    # values of the Waveform
                    values = self.waveform.get()
                    # add_slice option
                    if self.async_update.get() == "add_slice":
                        size = self.slice_size.get()
                        mod = len(values) % size
                        num_slices = len(values) // size + int(mod > 0)
                        for i in range(num_slices):
                            value_slice = values[i * size : min((i + 1) * size, len(values))]
                            logger.info(
                                f"Sending slice {i} of {self._slice_index} with length {len(value_slice)}"
                            )
                            self._run_subs(sub_type=self.SUB_MONITOR, value=value_slice)
                            self._send_async_update(index=self._slice_index, value=value_slice)
                            if self.delay_slice_update is True:
                                time.sleep(0.025)  # 25ms to be really fast
                            if self.stopped:
                                raise DeviceStopError(f"{self.name} was stopped")
                        self._slice_index += 1
                    # option add
                    elif self.async_update.get() == "add":
                        self._run_subs(sub_type=self.SUB_MONITOR, value=values)
                        self._send_async_update(value=values)
                    else:
                        # This should never happen, but just in case
                        # we raise an exception
                        raise ValueError(f"Invalid async_update type: {self.async_update.get()}")
                    if self.stopped:
                        raise DeviceStopError(f"{self.name} was stopped")
                status.set_finished()
            # pylint: disable=broad-except
            except Exception as exc:
                content = traceback.format_exc()
                status.set_exception(exc=exc)
                logger.warning(f"Error in {self.name} trigger; Traceback: {content}")

        self._trigger_thread = threading.Thread(target=acquire, args=(status,), daemon=True)
        self._trigger_thread.start()
        self._trigger_received += 1
        self.progress.put(
            value=self._trigger_received,
            max_value=self.frames.get(),
            done=self.frames.get() == self._trigger_received,
        )
        return status

    def _send_async_update(self, value: Any, index: int | None = None) -> None:
        """
        Send the async update to BEC.

        Args:
            index (int | None): The index of the slice to be sent. If None, the entire waveform is sent.
            value (Any): The value to be sent.
        """
        async_update_type = self.async_update.get()
        waveform_shape = self.waveform_shape.get()
        if async_update_type == "add_slice":
            if index is not None:
                async_update = {
                    "type": "add_slice",
                    "index": index,
                    "max_shape": [None, waveform_shape],
                }
            else:
                async_update = {"type": "add", "max_shape": [None, waveform_shape]}
        elif async_update_type == "add":
            async_update = {"type": "add", "max_shape": [None]}
        else:
            raise ValueError(
                f"Invalid async_update type: {async_update_type} for device {self.name}"
            )

        # TODO remove once BEC e2e test async data is updated to use AsyncSignal 'data'
        msg = messages.DeviceMessage(
            signals={self.waveform.name: {"value": value, "timestamp": time.time()}},
            metadata={"async_update": async_update},
        )
        # Send the message to BEC
        self.connector.xadd(
            MessageEndpoints.device_async_readback(
                scan_id=self.scan_info.msg.scan_id, device=self.name
            ),
            {"data": msg},
            expire=self._stream_ttl,
        )
        self.data.put(value, async_update=async_update)
        self.async_multi_data.put(
            {"data1": {"value": value}, "data2": {"value": value}}, async_update=async_update
        )

    def stage(self) -> list[object]:
        """Stage the camera for upcoming scan

        This method is called from BEC in preparation of a scan.
        It receives metadata about the scan from BEC,
        compiles it and prepares the camera for the scan.

        FYI: No data is written to disk in the simulation, but upon each trigger it
        is published to the device_monitor endpoint in REDIS.
        """
        if self._staged is Staged.yes:
            return super().stage()
        self.file_path.set(
            os.path.join(
                self.file_path.get(), self.file_pattern.get().format(self.scan_info.msg.scan_number)
            )
        )
        self.frames.set(
            self.scan_info.msg.num_points * self.scan_info.msg.scan_parameters["frames_per_trigger"]
        )
        self.exp_time.set(self.scan_info.msg.scan_parameters["exp_time"])
        self.burst.set(self.scan_info.msg.scan_parameters["frames_per_trigger"])
        self.stopped = False
        self._slice_index = 0
        self._trigger_received = 0
        logger.warning(f"Staged {self.name}, scan_id : {self.scan_info.msg.scan_id}")
        return super().stage()

    def unstage(self) -> list[object]:
        """Unstage the device

        Send reads from all config signals to redis
        """
        logger.warning(f"Unstaging {self.name}, {self._staged}")
        self._slice_index = 0
        if self.stopped is True or not self._staged:
            return super().unstage()
        return super().unstage()

    def stop(self, *, success=False):
        """Stop the device"""
        self.stopped = True
        if self._trigger_thread:
            self._trigger_thread.join()
        self._trigger_thread = None
        super().stop(success=success)


if __name__ == "__main__":  # pragma: no cover
    waveform = SimWaveform(name="waveform")
    waveform.sim.select_model("GaussianModel")



"""
Module for the optics shutter device at PSI beamlines.

Example config:
shutter:
    description: Optics Shutter A
    deviceClass: ophyd_devices.optics_shutter.OpticsShutter
    deviceConfig: {prefix: 'X10SA-EH1-PSYS:SH-A-'}
    enabled: true
    onFailure: retry
    readOnly: false
    readoutPriority: baseline
    softwareTrigger: false
    userParameter: {}
"""

from enum import IntEnum

from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal, EpicsSignalRO, Kind, Signal

from ophyd_devices import CompareStatus


class ShutterOpenState(IntEnum):
    """Enum for shutter open state."""

    OPEN = 1
    CLOSED = 0


class ShutterEnabled(IntEnum):
    """Enum for shutter enabled state."""

    ENABLED = 1
    DISABLED = 0


class ShutterControl(Device):
    """Control interface for the PVs related to the shutter."""

    is_open = Cpt(EpicsSignalRO, "OPEN", kind=Kind.config, auto_monitor=True)
    is_closed = Cpt(EpicsSignalRO, "CLOSE", kind=Kind.omitted)
    is_enabled = Cpt(EpicsSignalRO, "ENABLE", kind=Kind.config, auto_monitor=True)
    is_ok = Cpt(EpicsSignalRO, "OK", kind=Kind.omitted, auto_monitor=True)
    alarm = Cpt(EpicsSignalRO, "ALARM", kind=Kind.omitted, auto_monitor=True)
    set_open = Cpt(EpicsSignal, "OPEN-SET", kind=Kind.omitted)
    set_closed = Cpt(EpicsSignal, "CLOSE-SET", kind=Kind.omitted)


class ShutterOpenSignal(Signal):
    """
    ShutterOpenSignal. When called with 1, it will try to open the shutter.
    If called with 0, it will try to close the shutter.
    """

    def enabled(self) -> ShutterEnabled:
        """Check if the shutter is enabled."""
        return ShutterEnabled(self.parent.epics_control.is_enabled.get())

    def _check_enabled(self):
        if self.enabled() != ShutterEnabled.ENABLED:
            dev_name = self.parent.name if self.parent else "unknown"
            raise RuntimeError(f"The shutter {dev_name} is disabled!")

    def get(self) -> int:
        return self.parent.epics_control.is_open.get()

    def put(self, value: int, **kwargs) -> None:
        self._check_enabled()
        if value == ShutterOpenState.OPEN:
            self.parent.epics_control.set_open.put(1, **kwargs)
        elif value == ShutterOpenState.CLOSED:
            self.parent.epics_control.set_closed.put(1, **kwargs)
        else:
            raise ValueError("Invalid value for ShutterOpenSignal. Use 0 (CLOSED) or 1 (OPEN).")

    def set(self, value: int, **kwargs) -> CompareStatus:
        self.put(value, **kwargs)
        return CompareStatus(self.parent.epics_control.is_open, value)


class OpticsShutter(Device):
    """A shutter device with shutter open signal, and sub-device with full control PVs of the Epics implementation."""

    set_open = Cpt(
        ShutterOpenSignal,
        name="open",
        kind=Kind.omitted,
        doc="Signal to open/close the shutter. Use 0 (CLOSED) or 1 (OPEN).",
    )
    is_open = Cpt(
        EpicsSignalRO,
        "OPEN",
        kind=Kind.normal,
        auto_monitor=True,
        doc="Readback of the shutter open state. 0 (CLOSED) or 1 (OPEN).",
    )
    epics_control = Cpt(ShutterControl, suffix="", name="epics_control", kind=Kind.omitted)


if __name__ == "__main__":
    prefix = "X10SA-EH1-PSYS:SH-A-"
    print(f"Testing shutter device with prefix {prefix}")
    shutter = OpticsShutter(name="shutter", prefix=prefix)
    shutter.wait_for_connection()
    print(shutter.read())



import pytest

from ophyd_devices.devices.optics_shutter import OpticsShutter, ShutterEnabled, ShutterOpenState
from ophyd_devices.tests.utils import patched_device


@pytest.fixture(scope="function")
def mock_shutter():
    with patched_device(OpticsShutter, name="shutter", prefix="X10SA-EH1-PSYS:SH-A-") as shutter:
        yield shutter


def test_shutter_open(mock_shutter):
    mock_shutter.epics_control.is_enabled._read_pv.mock_data = ShutterEnabled.ENABLED.value
    mock_shutter.epics_control.is_open._read_pv.mock_data = ShutterOpenState.CLOSED.value
    mock_shutter.is_open._read_pv.mock_data = ShutterOpenState.CLOSED.value
    st = mock_shutter.set_open.set(1)
    assert not st.done
    mock_shutter.is_open._read_pv.mock_data = ShutterOpenState.OPEN.value
    mock_shutter.epics_control.is_open._read_pv.mock_data = ShutterOpenState.OPEN.value
    assert st.done


def test_shutter_close(mock_shutter):
    mock_shutter.epics_control.is_enabled._read_pv.mock_data = ShutterEnabled.ENABLED.value
    mock_shutter.epics_control.is_open._read_pv.mock_data = ShutterOpenState.OPEN.value
    mock_shutter.is_open._read_pv.mock_data = ShutterOpenState.OPEN.value
    st = mock_shutter.set_open.set(0)
    assert not st.done
    mock_shutter.is_open._read_pv.mock_data = ShutterOpenState.CLOSED.value
    mock_shutter.epics_control.is_open._read_pv.mock_data = ShutterOpenState.CLOSED.value
    assert st.done


def test_shutter_not_enabled(mock_shutter):
    with pytest.raises(RuntimeError) as e:
        mock_shutter.set_open.set(1)
    assert e.match(f"The shutter {mock_shutter.name} is disabled!")


def test_shutter_status(mock_shutter):
    mock_shutter.epics_control.is_open._read_pv.mock_data = ShutterOpenState.OPEN.value
    mock_shutter.is_open._read_pv.mock_data = ShutterOpenState.OPEN.value
    assert mock_shutter.is_open.get() == ShutterOpenState.OPEN.value
    mock_shutter.epics_control.is_open._read_pv.mock_data = ShutterOpenState.CLOSED.value
    mock_shutter.is_open._read_pv.mock_data = ShutterOpenState.CLOSED.value
    assert mock_shutter.is_open.get() == ShutterOpenState.CLOSED.value



from ophyd import Component as Cpt
from ophyd import EpicsSignal

from ophyd_devices.interfaces.base_classes.psi_positioner_base import PSISimplePositionerBase


class PSISimplePositioner(PSISimplePositionerBase):
    user_readback = Cpt(EpicsSignal, "R")
    user_setpoint = Cpt(EpicsSignal, "S")
    velocity = Cpt(EpicsSignal, "V")
    motor_done_move = Cpt(EpicsSignal, "D")


if __name__ == "__main__":  # pragma: no cover
    """You can use this to test against e.g. the Caproto fake motor IOC"""
    suffixes = {
        "user_readback": ".RBV",
        "user_setpoint": ".VAL",
        "velocity": ".VELO",
        "motor_done_move": ".DMOV",
    }
    pos = PSISimplePositioner(name="test", prefix="SIM:MOTOR", override_suffixes=suffixes)
    pos.wait_for_connection()
    st = pos.move(5, wait=False)



import abc
import functools
import socket
import time
import typing

import numpy as np
from bec_lib import bec_logger
from ophyd import Signal
from ophyd.utils.errors import DisconnectedError

logger = bec_logger.logger
# logger = bec_logger.logger("socket")


def raise_if_disconnected(fcn):
    """Decorator to catch attempted access to disconnected Galil channels."""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        if self.connected:
            return fcn(self, *args, **kwargs)
        raise DisconnectedError(f"{self.name} is not connected")

    return wrapper


DEFAULT_EPICSSIGNAL_VALUE = object()

_type_map = {
    "number": (float, np.floating),
    "array": (np.ndarray, list, tuple),
    "string": (str,),
    "integer": (int, np.integer),
}


def data_shape(val):
    """Determine data-shape (dimensions)

    Returns
    -------
    list
        Empty list if val is number or string, otherwise
        ``list(np.ndarray.shape)``
    """
    if data_type(val) != "array":
        return []

    try:
        return list(val.shape)
    except AttributeError:
        return [len(val)]


def data_type(val):
    """Determine the JSON-friendly type name given a value

    Returns
    -------
    str
        One of {'number', 'integer', 'array', 'string'}

    Raises
    ------
    ValueError if the type is not recognized
    """
    bad_iterables = (str, bytes, dict)
    if isinstance(val, typing.Iterable) and not isinstance(val, bad_iterables):
        return "array"

    for json_type, py_types in _type_map.items():
        if isinstance(val, py_types):
            return json_type

    raise ValueError(
        f"Cannot determine the appropriate bluesky-friendly data type for "
        f"value {val} of Python type {type(val)}. "
        f"Supported types include: int, float, str, and iterables such as "
        f"list, tuple, np.ndarray, and so on."
    )


class SocketSignal(abc.ABC, Signal):
    """
    A base class for signals that interact with a socket. This class provides the necessary structure
    for reading from and writing to a socket, as well as handling callbacks and subscriptions in a way
    that prevents recursive calls. Children should implement the `_socket_get` and `_socket_set` methods
    to define the specific logic for interacting with the socket. `get` and `put` methods should not be
    overridden by children to ensure proper callback handling.

    Args:
        name (str): The name of the signal.
        auto_monitor (bool, optional): If True, it indicates that the signal should be auto_monitored.
    """

    SUB_SETPOINT = "setpoint"

    def __init__(self, *, name, auto_monitor=False, **kwargs):

        super().__init__(name=name, **kwargs)
        self._auto_monitor = auto_monitor
        self._active_socket_callbacks: set[str] = set()

    @abc.abstractmethod
    def _socket_get(self): ...

    @abc.abstractmethod
    def _socket_set(self, val): ...

    def get(self, **kwargs):
        """
        Get the current value of the signal. Children should never
        override this method, but should implement the socket read logic in
        'socket_get' to ensure proper callback handling and caching of values.
        """
        if self.SUB_VALUE in self._active_socket_callbacks:
            return self._readback
        old_value = self._readback
        self._readback = self._socket_get()
        timestamp = time.time()
        self._metadata["timestamp"] = timestamp
        self._run_subs(
            sub_type=self.SUB_VALUE, old_value=old_value, value=self._readback, timestamp=timestamp
        )
        return self._readback

    def _run_subs(self, *args, sub_type, **kwargs):
        """
        This method runs the callbacks for a given subscription type. It is overridden to ensure that
        callbacks for the same subscription type can not trigger additional subscriptions of the same type.
        We thereby avoid that callbacks can triggered recursively. In practice, a callback may call 'get'
        or 'read' itself, but it won't trigger any recursive calls of the callbacks for the same subscription type.

        Args:
            sub_type (str): The subscription type for which to run the callbacks.
        """
        if sub_type in self._active_socket_callbacks:
            return
        try:
            self._active_socket_callbacks.add(sub_type)
            super()._run_subs(*args, sub_type=sub_type, **kwargs)
        finally:
            if sub_type in self._active_socket_callbacks:
                self._active_socket_callbacks.remove(sub_type)

    def put(self, value, connection_timeout=1, **kwargs):
        """
        Put method to send values to the socket. Children should never override this method,
        but should implement instead the socket write logic in 'socket_set' to ensure proper
        callback handling.

        Args:
            value (any): The value to set
            connection_timeout (float, optional): If not already connected, allow up to `connection_timeout` seconds
                for the connection to complete.
        """

        self.wait_for_connection(timeout=connection_timeout)
        old_value = self._readback
        self._socket_set(value)
        timestamp = time.time()
        # Super().put(..) triggers the SUB_VALUE callbacks with the new value.
        super().put(value, timestamp=timestamp, force=True)
        self._run_subs(
            sub_type=self.SUB_SETPOINT, old_value=old_value, value=value, timestamp=timestamp
        )

    def describe(self):
        """Provide schema and meta-data for :meth:`~BlueskyInterface.read`

        This keys in the `OrderedDict` this method returns must match the
        keys in the `OrderedDict` return by :meth:`~BlueskyInterface.read`.

        This provides schema related information, (ex shape, dtype), the
        source (ex PV name), and if available, units, limits, precision etc.

        Returns
        -------
        data_keys : OrderedDict
            The keys must be strings and the values must be dict-like
            with the ``event_model.event_descriptor.data_key`` schema.
        """
        if self._readback is DEFAULT_EPICSSIGNAL_VALUE:
            val = self.get()
        else:
            val = self._readback
        controller = self.root.controller if hasattr(self.root, "controller") else None
        return {
            self.name: {
                "source": f"{controller.name}:{self.name}" if controller else self.name,
                "dtype": data_type(val),
                "shape": data_shape(val),
            }
        }


class SocketIO:
    """SocketIO helper class for TCP IP connections"""

    def __init__(self, host: str, port: int, socket_timeout: int = 2):
        self.host = host
        self.port = port
        self.is_open = False
        self.socket_timeout = socket_timeout
        self._initialize_socket()

    def connect(self, timeout: int = 10):
        """
        Establish socket connection to host:port within timeout period

        Args:
            timeout (int): Time in seconds to wait for connection
        """
        logger.info(f"Connecting to {self.host}:{self.port}.")
        start_time = time.time()
        while time.time() - start_time < timeout:
            try:
                if self.sock is None:
                    self._initialize_socket()
                self.sock.connect((self.host, self.port))
                break
            except Exception as exc:
                self.sock = None
                logger.warning(
                    f"Connection to {self.host}:{self.port} failed after {time.time()-start_time:.2f} seconds"
                    f" with exception: {exc}. Retrying after 1 second..."
                )
                time.sleep(1)
        else:
            raise ConnectionError(
                f"Could not connect to {self.host}:{self.port} within {time.time()-start_time:.2f} seconds"
            )

    def _put(self, msg_bytes):
        logger.debug(f"put message: {msg_bytes}")
        return self.sock.send(msg_bytes)

    def _recv(self, buffer_length=1024):
        msg = self.sock.recv(buffer_length)
        logger.debug(f"recv message: {msg}")
        return msg

    def _initialize_socket(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout(self.socket_timeout)

    def put(self, msg):
        return self._put(msg)

    def receive(self, buffer_length=1024):
        return self._recv(buffer_length=buffer_length)

    def open(self, timeout: int = 10):
        """
        Open the socket connection to the host:port

        Args:
            timeout (int): Time in seconds to wait for connection
        """
        self.connect(timeout=timeout)
        self.is_open = True

    def close(self):
        self.sock.close()
        self.sock = None
        self.is_open = False


class SocketMock:
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.buffer_put = b""
        self.buffer_recv = [b" -12800"]
        self.is_open = False
        # self.open()

    def connect(self, timeout: int = 10):
        print(f"connecting to {self.host} port {self.port}")

    def _put(self, msg_bytes):
        self.buffer_put = msg_bytes
        print(self.buffer_put)

    def _recv(self, buffer_length=1024):
        print(self.buffer_recv)
        if isinstance(self.buffer_recv, list):
            if len(self.buffer_recv) > 0:
                ret_val = self.buffer_recv.pop(0)
            else:
                ret_val = b""
            return ret_val
        return self.buffer_recv

    def _initialize_socket(self):
        pass

    def put(self, msg):
        return self._put(msg)

    def receive(self, buffer_length=1024):
        return self._recv(buffer_length=buffer_length)

    def open(self, timeout: int = 10):
        self._initialize_socket()
        self.is_open = True

    def close(self):
        self.sock = None
        self.is_open = False



"""Module for simulated positioner devices."""

import threading
import time as ttime
import traceback

import numpy as np
from bec_lib.logger import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, DeviceStatus, Kind, PositionerBase
from ophyd.utils import LimitError
from typeguard import typechecked

from ophyd_devices.sim.sim_data import SimulatedPositioner
from ophyd_devices.sim.sim_signals import ReadOnlySignal, SetableSignal
from ophyd_devices.sim.sim_utils import LinearTrajectory, stop_trajectory
from ophyd_devices.utils.errors import DeviceStopError

logger = bec_logger.logger


class SimPositioner(Device, PositionerBase):
    """
    A simulated device mimicking any 1D Axis device (position, temperature, rotation).

    >>> motor = SimPositioner(name="motor")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argument, passed on to all signals of the device.\
    Optional parameters:
    ----------
    delay (int)             : If 0, execution of move will be instant. If 1, execution will depend on motor velocity. Default is 1.
    update_frequency (int)  : Frequency in Hz of the update of the simulated state during a move. Default is 2 Hz.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    limits (tuple)          : Tuple of the low and high limits of the positioner. Overrides low/high_limit_travel is specified. Default is None.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation, device_manager is passed on automatically.
    sim_init (dict)         : Dictionary to initiate parameters of the simulation, check simulation type defaults for more details.

    """

    # Specify which attributes are accessible via BEC client
    USER_ACCESS = ["sim", "readback", "registered_proxies"]

    sim_cls = SimulatedPositioner

    # Define the signals as class attributes
    readback = Cpt(ReadOnlySignal, name="readback", value=0, kind=Kind.hinted)
    setpoint = Cpt(SetableSignal, value=0, kind=Kind.normal)
    motor_is_moving = Cpt(SetableSignal, value=0, kind=Kind.normal)

    # Config signals
    velocity = Cpt(SetableSignal, value=100, kind=Kind.config)
    acceleration = Cpt(SetableSignal, value=1, kind=Kind.config)
    tolerance = Cpt(SetableSignal, value=0.5, kind=Kind.config)

    # Ommitted signals
    high_limit_travel = Cpt(SetableSignal, value=0, kind=Kind.omitted)
    low_limit_travel = Cpt(SetableSignal, value=0, kind=Kind.omitted)
    unused = Cpt(SetableSignal, value=1, kind=Kind.omitted)

    SUB_READBACK = "readback"
    _default_sub = SUB_READBACK

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name,
        *,
        delay: int = 1,
        update_frequency=2,
        precision=3,
        limits=None,
        parent=None,
        kind=None,
        device_manager=None,
        sim_init: dict = None,
        **kwargs,
    ):
        self.move_thread = None
        self.delay = delay
        self.device_manager = device_manager
        self.precision = precision
        self.sim_init = sim_init
        self._registered_proxies = {}

        self.update_frequency = update_frequency
        self._stopped = False

        self.sim = self.sim_cls(parent=self, **kwargs)
        self._status_list = []

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self.sim.sim_state[self.name] = self.sim.sim_state.pop(self.readback.name, None)
        self.readback.name = self.name
        if limits is not None:
            assert len(limits) == 2
            self.low_limit_travel.put(limits[0])
            self.high_limit_travel.put(limits[1])
        if self.sim_init:
            self.sim.set_init(self.sim_init)

    @property
    def limits(self):
        """Return the limits of the simulated device."""
        return (self.low_limit_travel.get(), self.high_limit_travel.get())

    @property
    def low_limit(self):
        """Return the low limit of the simulated device."""
        return self.limits[0]

    @property
    def high_limit(self):
        """Return the high limit of the simulated device."""
        return self.limits[1]

    def registered_proxies(self) -> None:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies

    # pylint: disable=arguments-differ
    def check_value(self, value: any):
        """
        Check that requested position is within existing limits.

        This function has to be implemented on the top level of the positioner.
        """
        low_limit, high_limit = self.limits

        if low_limit < high_limit and not low_limit <= value <= high_limit:
            raise LimitError(f"position={value} not within limits {self.limits}")

    @typechecked
    def _set_sim_state(self, signal_name: str, value: any) -> None:
        """Update the simulated state of the device."""
        self.sim.sim_state[signal_name]["value"] = value
        self.sim.sim_state[signal_name]["timestamp"] = ttime.time()

    def _get_sim_state(self, signal_name: str) -> any:
        """Return the simulated state of the device."""
        return self.sim.sim_state[signal_name]["value"]

    def _update_state(self, val):
        """Update the state of the simulated device."""
        old_readback = self._get_sim_state(self.readback.name)
        self._set_sim_state(self.readback.name, val)

        # Run subscription on "readback"
        self._run_subs(
            sub_type=self.SUB_READBACK,
            old_value=old_readback,
            value=self.sim.sim_state[self.readback.name]["value"],
            timestamp=self.sim.sim_state[self.readback.name]["timestamp"],
        )

    def _move_to_setpoint(self) -> None:
        """Move the simulated device to the setpoint."""
        try:
            while True:
                setpoint = self.setpoint.get()
                value = self.readback.get()

                increment = np.sign(setpoint - value) * self.velocity.get() / self.update_frequency
                next_val = value + increment + np.random.uniform(-1, 1) * self.tolerance.get()

                # Check if next_val would overshoot the setpoint
                if (increment > 0 and next_val > setpoint) or (
                    increment < 0 and next_val < setpoint
                ):
                    next_val = setpoint + np.random.uniform(-1, 1) * self.tolerance.get()

                self._update_state(next_val)
                if np.isclose(setpoint, next_val, atol=self.tolerance.get()):
                    break
                if self._stopped:
                    raise DeviceStopError(f"{self.name} was stopped")
                ttime.sleep(1 / self.update_frequency)
            self._update_state(self.readback.get())
            for status in self._status_list:
                status.set_finished()
        # pylint: disable=broad-except
        except Exception as exc:
            content = traceback.format_exc()
            logger.warning(
                f"Error in on_complete call in device {self.name}. Error traceback: {content}"
            )
            for status in self._status_list:
                status.set_exception(exc=exc)
        finally:
            self.motor_is_moving.put(0)
            if not self._stopped:
                self._update_state(self.readback.get())
            self._status_list = []

    def move(self, value: float, **kwargs) -> DeviceStatus:
        """Change the setpoint of the simulated device, and simultaneously initiate a motion."""
        self._stopped = False
        self.check_value(value)
        self.motor_is_moving.put(1)
        self.setpoint.put(value)

        st = DeviceStatus(device=self)
        self._status_list.append(st)
        if self.delay:
            if self.move_thread is None or not self.move_thread.is_alive():
                self.move_thread = threading.Thread(target=self._move_to_setpoint)
                self.move_thread.start()
        else:
            self._done_moving()
            self.motor_is_moving.put(0)
            self._update_state(value)
            st.set_finished()
        return st

    def stop(self, *, success=False):
        """Stop the motion of the simulated device."""
        self._stopped = True
        if self.move_thread:
            self.move_thread.join()
        self.move_thread = None
        super().stop(success=success)

    @property
    def position(self) -> float:
        """Return the current position of the simulated device."""
        return self.readback.get()

    @property
    def egu(self):
        """Return the engineering units of the simulated device."""
        return "mm"


class SimLinearTrajectoryPositioner(SimPositioner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _move_and_finish(self, start_pos, end_pos, st):
        acc_time = (
            self.acceleration.get()
        )  # acceleration in Ophyd refers to acceleration time in seconds
        vel = self.velocity.get()
        acc = abs(vel / acc_time)
        traj = LinearTrajectory(start_pos, end_pos, vel, acc)

        try:
            while not traj.ended:
                ttime.sleep(1 / self.update_frequency)
                self._update_state(traj.position())
                if self._stopped:
                    # simulate deceleration
                    traj = stop_trajectory(traj)
                    while not traj.ended:
                        ttime.sleep(1 / self.update_frequency)
                        self._update_state(traj.position())
                    raise DeviceStopError(f"{self.name} was stopped")
            st.set_finished()
        # pylint: disable=broad-except
        except Exception as exc:
            content = traceback.format_exc()
            logger.warning(
                f"Error in on_complete call in device {self.name}. Error traceback: {content}"
            )
            st.set_exception(exc=exc)
        finally:
            self._set_sim_state(self.motor_is_moving.name, 0)

    def move(self, value: float, **kwargs) -> DeviceStatus:
        """Change the setpoint of the simulated device, and simultaneously initiate a motion."""
        self._stopped = False
        self.check_value(value)
        self._set_sim_state(self.motor_is_moving.name, 1)
        self._set_sim_state(self.setpoint.name, value)

        st = DeviceStatus(device=self)
        if self.delay:
            if self.move_thread is None or not self.move_thread.is_alive():
                self.move_thread = threading.Thread(
                    target=self._move_and_finish, args=(self.position, value, st)
                )
                self.move_thread.start()
            else:
                raise RuntimeError(f"{self.name} is already moving. Cannot start a new move.")
        else:
            self._done_moving()
            self._set_sim_state(self.motor_is_moving.name, 0)
            self._update_state(value)
            st.set_finished()
        return st



# pylint: skip-file
import time
from unittest import mock

import pytest
from ophyd import DeviceStatus, Staged
from ophyd.utils.errors import RedundantStaging

from ophyd_devices.interfaces.base_classes.psi_device_base import PSIDeviceBase
from ophyd_devices.utils.errors import DeviceStopError, DeviceTimeoutError


@pytest.fixture
def detector_base():
    yield PSIDeviceBase(name="test_detector")


def test_detector_base_init(detector_base):
    assert detector_base.stopped is False
    assert detector_base.name == "test_detector"
    assert detector_base.staged == Staged.no
    assert detector_base.destroyed == False


def test_stage(detector_base):
    assert detector_base._staged == Staged.no
    assert detector_base.stopped is False
    detector_base._staged = Staged.no
    with mock.patch.object(detector_base, "on_stage") as mock_on_stage:
        rtr = detector_base.stage()
        assert isinstance(rtr, list)
        assert mock_on_stage.called is True
        with pytest.raises(RedundantStaging):
            detector_base.stage()
        detector_base._staged = Staged.no
        detector_base.stopped = True
        detector_base.stage()
        assert detector_base.stopped is False
        assert mock_on_stage.call_count == 2


def test_pre_scan(detector_base):
    with mock.patch.object(detector_base, "on_pre_scan") as mock_on_pre_scan:
        detector_base.pre_scan()
        mock_on_pre_scan.assert_called_once()


def test_trigger(detector_base):
    status = DeviceStatus(detector_base)
    with mock.patch.object(
        detector_base, "on_trigger", side_effect=[None, status]
    ) as mock_on_trigger:
        st = detector_base.trigger()
        assert isinstance(st, DeviceStatus)
        st.wait(timeout=1)
        assert st.done is True
        st = detector_base.trigger()
        assert st.done is False
        assert id(st) == id(status)


def test_unstage(detector_base):
    detector_base.stopped = True
    with (mock.patch.object(detector_base, "on_unstage") as mock_on_unstage,):
        rtr = detector_base.unstage()
        assert isinstance(rtr, list)
        assert mock_on_unstage.call_count == 1
        detector_base.stopped = False
        rtr = detector_base.unstage()
        assert isinstance(rtr, list)
        assert mock_on_unstage.call_count == 2


def test_complete(detector_base):
    status = DeviceStatus(detector_base)
    with mock.patch.object(
        detector_base, "on_complete", side_effect=[None, status]
    ) as mock_on_complete:
        st = detector_base.complete()
        assert isinstance(st, DeviceStatus)
        time.sleep(0.1)
        assert st.done is True
        st = detector_base.complete()
        assert st.done is False
        assert id(st) == id(status)


def test_stop(detector_base):
    with mock.patch.object(detector_base, "on_stop") as mock_on_stop:
        detector_base.stop()
        mock_on_stop.assert_called_once()
        assert detector_base.stopped is True



from unittest import mock

from ophyd_devices.utils.controller import Controller


def test_controller_off(dm_with_devices):
    controller = Controller(
        socket_cls=mock.MagicMock(),
        socket_host="dummy",
        socket_port=123,
        device_manager=dm_with_devices,
    )
    controller.on()
    with mock.patch.object(controller.device_manager, "config_helper") as mock_config_helper:
        with mock.patch.object(controller.sock, "close") as mock_close:
            controller.off()
            assert controller.sock is None
            assert controller.connected is False
            mock_close.assert_called_once()

            # make sure it is indempotent
            controller.off()
    controller._reset_controller()


def test_controller_on(dm_with_devices):
    socket_cls = mock.MagicMock()
    Controller._controller_instances = {}
    controller = Controller(
        socket_cls=socket_cls, socket_host="dummy", socket_port=123, device_manager=dm_with_devices
    )
    controller.on()
    assert controller.sock is not None
    assert controller.connected is True
    socket_cls().open.assert_called_once()

    # make sure it is indempotent
    controller.on()
    socket_cls().open.assert_called_once()
    controller._reset_controller()


def test_controller_with_multiple_axes(dm_with_devices):
    """Test that turning the controller on and off enables/disables all axes attached to it."""
    socket_cls = mock.MagicMock()
    Controller._controller_instances = {}
    Controller._axes_per_controller = 2
    controller = Controller(
        socket_cls=socket_cls, socket_host="dummy", socket_port=123, device_manager=dm_with_devices
    )
    with mock.patch.object(controller.device_manager, "config_helper") as mock_config_helper:
        # Disable samx, samy first
        dm_with_devices.devices.get("samx").enabled = False
        dm_with_devices.devices.get("samy").enabled = False
        # Set axes on the controller
        controller.set_axis(axis=dm_with_devices.devices["samx"], axis_nr=0)
        controller.set_axis(axis=dm_with_devices.devices["samy"], axis_nr=1)
        # Turn the controller on, should turn the controller on, but not enable the axes
        controller.on()
        assert dm_with_devices.devices.get("samx").enabled is False
        assert dm_with_devices.devices.get("samy").enabled is False
        assert controller.connected is True
        controller.set_all_devices_enabled(True)
        assert dm_with_devices.devices.get("samx").enabled is True
        assert dm_with_devices.devices.get("samy").enabled is True
        # Disable one axis after another, the last one should turn the controller off
        controller.set_device_enabled("samx", False)
        assert controller.connected is True
        assert dm_with_devices.devices.get("samx").enabled is False
        assert dm_with_devices.devices.get("samy").enabled is True
        controller.set_device_enabled("samy", False)
        assert dm_with_devices.devices.get("samy").enabled is False
        # Enabling one axis should turn the controller back on
        assert controller.connected is False
        controller.set_device_enabled("samx", True)
        assert controller.connected is True


def test_remove_axis_from_controller(dm_with_devices):
    """Test removing an axis from a controller."""
    socket_cls = mock.MagicMock()
    Controller._controller_instances = {}
    Controller._axes_per_controller = 2
    controller = Controller(
        socket_cls=socket_cls, socket_host="dummy", socket_port=123, device_manager=dm_with_devices
    )
    # Set axes on the controller
    controller.set_axis(axis=dm_with_devices.devices["samx"], axis_nr=0)
    controller.set_axis(axis=dm_with_devices.devices["samy"], axis_nr=1)
    assert len(controller._axis) == 2
    with mock.patch.object(controller, "off") as mock_off:
        # Remove one axis
        controller.remove_axis(axis_nr=0)
        assert len([a for a in controller._axis if a is not None]) == 1
        assert "samy" in [a.name for a in controller._axis if a is not None]
        mock_off.assert_not_called()

        # Remove the last axis, should disable the controller
        controller.remove_axis(axis_nr=1)
        assert len([a for a in controller._axis if a is not None]) == 0
        assert controller.connected is False
        mock_off.assert_called_once_with(update_config=False)
